import numpy
from roguewave.interpolate.dataset import (
    interpolate_dataset_grid,
    interpolate_dataset_along_axis,
)
from roguewave.tools.math import wrapped_difference
from roguewave.tools.time import to_datetime64
from roguewave.wavetheory.lineardispersion import (
    inverse_intrinsic_dispersion_relation,
    intrinsic_group_velocity,
)
from roguewave.wavespectra.estimators import (
    estimate_directional_distribution,
    Estimators,
)
from ._tools import fill_zeros_or_nan_in_tail
from typing import Iterator, Hashable, TypeVar, Union, List, Literal, Mapping
from xarray import (
    Dataset,
    DataArray,
    open_dataset,
    concat,
    ones_like,
    where,
    zeros_like,
)
from xarray.core.coordinates import DatasetCoordinates
from warnings import warn

NAME_F: Literal["frequency"] = "frequency"
NAME_D: Literal["direction"] = "direction"
NAME_T: Literal["time"] = "time"
NAME_E: Literal["variance_density"] = "variance_density"
NAME_a1: Literal["a1"] = "a1"
NAME_b1: Literal["b1"] = "b1"
NAME_a2: Literal["a2"] = "a2"
NAME_b2: Literal["b2"] = "b2"
NAME_LAT: Literal["latitude"] = "latitude"
NAME_LON: Literal["longitude"] = "longitude"
NAME_DEPTH: Literal["depth"] = "depth"
NAMES_2D = (NAME_F, NAME_D, NAME_T, NAME_E, NAME_LAT, NAME_LON, NAME_DEPTH)
NAMES_1D = (
    NAME_F,
    NAME_T,
    NAME_E,
    NAME_LAT,
    NAME_LON,
    NAME_a1,
    NAME_b1,
    NAME_a2,
    NAME_b2,
    NAME_DEPTH,
)
SPECTRAL_VARS = (NAME_E, NAME_a1, NAME_b1, NAME_a2, NAME_b2)
SPECTRAL_DIMS = (NAME_F, NAME_D)
SPACE_TIME_DIMS = (NAME_T, NAME_LON, NAME_LAT)

_T = TypeVar("_T")


class DatasetWrapper:
    """
    A class that wraps a dataset object and passes through some of its primary
    functionality (get/set etc.). Used here mostly to make explicit what parts
    of the Dataset interface we actually expose in frequency objects. Note that
    we do not claim- or try to obtain completeness here. If full capabilities
    of the dataset object are needed we can simple operate directly on the
    dataset object itself.
    """

    def __init__(self, dataset: Dataset):
        self.dataset = dataset

    def __getitem__(self, item):
        return self.dataset.__getitem__(item)

    def __setitem__(self, key, value) -> None:
        return self.dataset.__setitem__(key, value)

    def __copy__(self: _T) -> _T:
        cls = self.__class__
        return cls(self.dataset.copy())

    def __len__(self):
        return len(self.dataset)

    def copy(self: _T, deep=True) -> _T:
        if deep:
            return self.__deepcopy__({})
        else:
            return self.__copy__()

    def __deepcopy__(self: _T, memodict) -> _T:
        cls = self.__class__
        return cls(self.dataset.copy(deep=True))

    def coords(self) -> DatasetCoordinates:
        return self.dataset.coords

    def keys(self):
        return self.dataset.keys()

    def __contains__(self, key: object) -> bool:
        return key in self.dataset

    def __iter__(self) -> Iterator[Hashable]:
        return self.dataset.__iter__()

    def sel(self: _T, *args, **kwargs) -> _T:
        cls = type(self)
        dataset = Dataset()
        for var in self.dataset:
            dataset = dataset.assign({var: self.dataset[var].sel(*args, **kwargs)})
        return cls(dataset=dataset)

    def isel(self: _T, *args, **kwargs) -> _T:
        cls = type(self)
        dataset = Dataset()
        for var in self.dataset:
            dataset = dataset.assign({var: self.dataset[var].isel(*args, **kwargs)})
        return cls(dataset=dataset)


class WaveSpectrum(DatasetWrapper):
    frequency_units = "Hertz"
    angular_units = "Degrees"
    spectral_density_units = "m**2/Hertz"
    angular_convention = (
        "Wave travel direction (going-to), " "measured anti-clockwise from East"
    )
    bulk_properties = (
        "m0",
        "hm0",
        "tm01",
        "tm02",
        "peak_period",
        "peak_direction",
        "peak_directional_spread",
        "mean_direction",
        "mean_directional_spread",
        "peak_frequency",
        "peak_wavenumber",
        "latitude",
        "longitude",
        "time",
    )

    def __init__(self, dataset: Dataset):
        super(WaveSpectrum, self).__init__(dataset)

    def __add__(self: _T, other: _T) -> _T:
        spectrum = self.copy(deep=True)
        spectrum.dataset[NAME_E] = spectrum.dataset[NAME_E] + other.dataset[NAME_E]
        return spectrum

    def __sub__(self: _T, other: _T) -> _T:
        spectrum = self.copy(deep=True)
        spectrum.dataset[NAME_E] = spectrum.dataset[NAME_E] - other.dataset[NAME_E]
        return spectrum

    def __neg__(self: _T) -> _T:
        """
        Negate self- that is -spectrum
        :return: spectrum with all spectral values taken to have the opposite
            sign.
        """
        spectrum = self.copy(deep=True)
        spectrum.dataset[NAME_E] = -spectrum.dataset[NAME_E]
        return spectrum

    def __len__(self):
        return self.number_of_spectra

    def __getitem__(self: _T, item) -> _T:

        if isinstance(item, tuple):
            if len(item) < self.ndims:
                raise ValueError(
                    f"Indexing requires same number of inputs as dimensions: {self.ndims}"
                )
            space_time_index = item[: -len(self.dims_spectral)]
        else:
            if not self.ndims == 1:
                raise ValueError(
                    f"Indexing requires same number of inputs as dimensions: {self.ndims}"
                )
            space_time_index = []

        dataset = Dataset()
        for var in self.dataset:
            if var in SPECTRAL_VARS:
                dataset = dataset.assign({var: self.dataset[var].__getitem__(item)})
            else:
                if space_time_index:
                    # array
                    dataset = dataset.assign(
                        {var: self.dataset[var].__getitem__(space_time_index)}
                    )
                else:
                    # Scalar
                    dataset = dataset.assign({var: self.dataset[var]})

        cls = type(self)
        return cls(dataset)

    @property
    def ndims(self):
        return len(self.dims)

    @property
    def frequency_step(self) -> DataArray:
        prepend = 2 * self.frequency[0] - self.frequency[1]
        append = 2 * self.frequency[-1] - self.frequency[-2]
        diff = numpy.diff(self.frequency, append=append, prepend=prepend)
        return DataArray(
            data=(diff[0:-1] * 0.5 + diff[1:] * 0.5),
            dims=NAME_F,
            coords={NAME_F: self.frequency},
        )

    def is_invalid(self) -> DataArray:
        return self.variance_density.isnull().all(dim=self.dims_spectral)

    def is_valid(self) -> DataArray:
        return ~self.is_invalid()

    def drop_invalid(self: _T) -> _T:
        return self._apply_filter(self.is_valid())

    def where(self: _T, condition) -> _T:
        return self._apply_filter(condition)

    def _apply_filter(self: _T, boolean_mask: DataArray) -> _T:
        dataset = Dataset()
        for var in self.dataset:
            data = self.dataset[var].where(
                boolean_mask.reindex_like(self.dataset[var]), drop=True
            )
            dataset = dataset.assign({var: data})

        cls = type(self)
        return cls(dataset)

    def mean(self: _T, dim, skipna=False) -> _T:
        """
        Calculate the mean value of the spectrum along the given dimension.
        :param dim: dimension to average over
        :param skipna: whether or not to "skip" nan values; if True behaves as numpy.nanmean
        :return:
        """
        if dim in SPECTRAL_DIMS:
            raise ValueError("Cannot calculate mean over spectral dimensions")

        cls = type(self)
        dataset = Dataset()
        # Todo: fix averaging over longitude for (prime/anti) meridian issues
        dataset = dataset.assign({dim: self.dataset[dim].mean(dim=dim, skipna=skipna)})
        for x in self.dataset:
            dataset = dataset.assign({x: self.dataset[x].mean(dim=dim, skipna=skipna)})
        return cls(dataset)

    def flatten(self: "WaveSpectrum", flattened_coordinate="linear_index") -> _T:
        """
        Serialize the non-spectral dimensions creating a single leading dimension without a coordinate.
        """

        # Get the current dimensions and shape
        dims = self.dims_space_time
        coords = self.coords_space_time
        shape = self.space_time_shape()
        length = numpy.product(shape)

        # Calculate the flattened shape
        new_shape = (length,)
        new_spectral_shape = (length, *self.spectral_shape())
        new_dims = [flattened_coordinate] + self.dims_spectral

        linear_index = DataArray(
            data=numpy.arange(0, length), dims=flattened_coordinate
        )
        indices = numpy.unravel_index(linear_index.values, shape)

        dataset = {}
        for index, dim in zip(indices, dims):
            dataset[dim] = DataArray(
                data=coords[dim].values[index], dims=flattened_coordinate
            )

        for name in self.dataset:
            if name in SPECTRAL_VARS:
                x = DataArray(
                    data=self.dataset[name].values.reshape(new_spectral_shape),
                    dims=new_dims,
                    coords=self.coords_spectral,
                )
            else:
                x = DataArray(
                    data=self.dataset[name].values.reshape(new_shape),
                    dims=flattened_coordinate,
                )
            dataset[name] = x

        cls = type(self)
        return cls(Dataset(dataset))

    def sum(self: _T, dim, skipna=False) -> _T:
        """
        Calculate the sum value of the spectrum along the given dimension.
        :param dim: dimension to sum over
        :param skipna: whether or not to "skip" nan values; if True behaves as numpy.nansum
        :return:
        """

        if dim in SPECTRAL_DIMS:
            raise ValueError("Cannot calculate sum over spectral dimensions")

        cls = type(self)
        dataset = Dataset()
        # we assign the average coordinate to the dimension we sum over
        dataset = dataset.assign({dim: self.dataset[dim].mean(dim=dim, skipna=skipna)})
        for x in self.dataset:
            dataset = dataset.assign({x: self.dataset[x].sum(dim=dim, skipna=skipna)})
        return cls(dataset)

    def std(self: _T, dim, skipna=False) -> _T:
        """
        Calculate the standard deviation of the spectrum along the given dimension.
        :param dim: dimension to calculate standard deviation over
        :param skipna: whether or not to "skip" nan values; if True behaves as numpy.nanstd
        :return:
        """
        if dim in SPECTRAL_DIMS:
            raise ValueError(
                "Cannot calculate standard deviation over spectral dimensions"
            )

        cls = type(self)
        dataset = Dataset()
        # we assign the average coordinate to the dimension we calculate the std over
        dataset = dataset.assign({dim: self.dataset[dim].mean(dim=dim, skipna=skipna)})
        for x in self.dataset:
            dataset = dataset.assign({x: self.dataset[x].std(dim=dim, skipna=skipna)})
        return cls(dataset)

    def shape(self):
        return self.variance_density.shape

    def spectral_shape(self):
        number_of_spectral_dims = len(self.dims_spectral)
        return self.shape()[-number_of_spectral_dims:]

    def space_time_shape(self):
        number_of_spectral_dims = len(self.dims_spectral)
        return self.shape()[:-number_of_spectral_dims]

    def frequency_moment(self, power: int, fmin=0, fmax=numpy.inf) -> DataArray:
        """
        Calculate a "frequency moment" over the given range. A frequency moment
        here refers to the integral:

                    Integral-over-frequency-range[ e(f) * f**power ]

        :param power: power of the frequency
        :param fmin: minimum frequency
        :param fmax: maximum frequency
        :return: frequency moment
        """
        _range = self._range(fmin, fmax)

        # Integrate dataset over frequencies. Make sure to fill any NaN entries with 0 before the integration.
        return (
            (self.e.isel({NAME_F: _range}) * self.frequency[_range] ** power)
            .fillna(0)
            .integrate(coord=NAME_F)
        )

    @property
    def number_of_frequencies(self) -> int:
        """
        :return: number of frequencies
        """
        return len(self.frequency)

    @property
    def dims_space_time(self) -> List[str]:
        return [str(x) for x in self.variance_density.dims if x not in (SPECTRAL_DIMS)]

    @property
    def coords_space_time(self) -> Mapping[str, DataArray]:
        return {dim: self.dataset[dim] for dim in self.dims_space_time}

    @property
    def coords_spectral(self) -> Mapping[str, DataArray]:
        return {dim: self.dataset[dim] for dim in self.dims_spectral}

    @property
    def dims_spectral(self) -> List[str]:
        return [str(x) for x in self.variance_density.dims if x in (SPECTRAL_DIMS)]

    @property
    def dims(self) -> List[str]:
        return [str(x) for x in self.variance_density.dims]

    @property
    def number_of_spectra(self):
        dims = self.dims_space_time
        if dims:
            shape = 1
            for d in dims:
                shape *= len(self.dataset[d])
            return shape
        else:
            return 1

    @property
    def spectral_values(self) -> DataArray:
        """
        :return: Spectral levels
        """
        return self.dataset[NAME_E]

    @property
    def radian_frequency(self) -> DataArray:
        """
        :return: Radian frequency
        """
        data_array = self.dataset[NAME_F] * 2 * numpy.pi
        data_array.name = "radian_frequency"
        return data_array

    @property
    def latitude(self) -> DataArray:
        """
        :return: latitudes
        """
        return self.dataset[NAME_LAT]

    @property
    def longitude(self) -> DataArray:
        """
        :return: longitudes
        """
        return self.dataset[NAME_LON]

    @property
    def time(self) -> DataArray:
        """
        :return: Time
        """
        return self.dataset[NAME_T]

    @property
    def variance_density(self) -> DataArray:
        """
        :return: Time
        """
        return self.dataset[NAME_E]

    @property
    def e(self) -> DataArray:
        """
        :return: 1D spectral values (directionally integrated spectrum).
            Equivalent to self.spectral_values if this is a 1D spectrum.
        """
        return self.dataset[NAME_E]

    @property
    def a1(self) -> DataArray:
        """
        :return: normalized Fourier moment cos(theta)
        """
        return self.dataset[NAME_a1]

    @property
    def b1(self) -> DataArray:
        """
        :return: normalized Fourier moment sin(theta)
        """
        return self.dataset[NAME_b1]

    @property
    def a2(self) -> DataArray:
        """
        :return: normalized Fourier moment cos(2*theta)
        """
        return self.dataset[NAME_a2]

    @property
    def b2(self) -> DataArray:
        """
        :return: normalized Fourier moment sin(2*theta)
        """
        return self.dataset[NAME_b2]

    @property
    def A1(self) -> DataArray:
        """
        :return: Fourier moment cos(theta)
        """
        return self.a1 * self.e

    @property
    def B1(self) -> DataArray:
        """
        :return: Fourier moment sin(theta)
        """
        return self.b1 * self.e

    @property
    def A2(self) -> DataArray:
        """
        :return: Fourier moment cos(2*theta)
        """
        return self.a2 * self.e

    @property
    def B2(self) -> DataArray:
        """
        :return: Fourier moment sin(2*theta)
        """
        return self.b2 * self.e

    @property
    def frequency(self) -> DataArray:
        """
        :return: Frequencies (Hz)
        """
        return self.dataset[NAME_F]

    def m0(self, fmin=0, fmax=numpy.inf) -> DataArray:
        """
        Zero order frequency moment. Also referred to as variance or energy.

        :param fmin: minimum frequency
        :param fmax: maximum frequency
        :return: variance/energy
        """
        return self.frequency_moment(0, fmin, fmax)

    def m1(self, fmin=0, fmax=numpy.inf) -> DataArray:
        """
        First order frequency moment. Primarily used in calculating a mean
        period measure (Tm01)

        :param fmin: minimum frequency
        :param fmax: maximum frequency
        :return: first order frequency moment.
        """
        return self.frequency_moment(1, fmin, fmax)

    def wave_speed(self) -> DataArray:
        """
        :return:
        """

        # Note we multiply inverse wavenumber with frequency to force xarray to return a number_of_points by
        # by number of frequencies data structure.
        return (1 / self.wavenumber) * self.radian_frequency

    def peak_wave_speed(self) -> DataArray:
        return 2 * numpy.pi * self.peak_frequency() / self.peak_wavenumber

    def m2(self, fmin=0, fmax=numpy.inf) -> DataArray:
        """
        Second order frequency moment. Primarily used in calculating the zero
        crossing period (Tm02)

        :param fmin: minimum frequency
        :param fmax: maximum frequency
        :return: Second order frequency moment.
        """
        return self.frequency_moment(2, fmin, fmax)

    def hm0(self, fmin=0, fmax=numpy.inf) -> DataArray:
        """
        Significant wave height estimated from the spectrum, i.e. waveheight
        h estimated from variance m0. Common notation in literature.

        :param fmin: minimum frequency
        :param fmax: maximum frequency
        :return: Significant wave height
        """
        return 4 * numpy.sqrt(self.m0(fmin, fmax))

    def tm01(self, fmin=0, fmax=numpy.inf) -> DataArray:
        """
        Mean period, estimated as the inverse of the center of mass of the
        spectral curve under the 1d spectrum.

        :param fmin: minimum frequency
        :param fmax: maximum frequency
        :return: Mean period
        """
        return self.m0(fmin, fmax) / self.m1(fmin, fmax)

    def tm02(self, fmin=0, fmax=numpy.inf) -> DataArray:
        """
        Zero crossing period based on Rice's spectral estimate.

        :param fmin: minimum frequency
        :param fmax: maximum frequency
        :return: Zero crossing period
        """
        return numpy.sqrt(self.m0(fmin, fmax) / self.m2(fmin, fmax))

    def peak_index(self, fmin=0, fmax=numpy.inf) -> DataArray:
        """
        Index of the peak frequency of the 1d spectrum within the given range
        :param fmin: minimum frequency
        :param fmax: maximum frequency
        :return: peak indices
        """
        return self.e.where(self._range(fmin, fmax), 0).argmax(dim=NAME_F)

    def peak_frequency(self, fmin=0, fmax=numpy.inf) -> DataArray:
        """
        Peak frequency of the spectrum, i.e. frequency at which the spectrum
        obtains its maximum.

        :param fmin: minimum frequency
        :param fmax: maximum frequency
        :return: peak frequency
        """
        return self.dataset[NAME_F][self.peak_index(fmin, fmax)]

    def peak_angular_frequency(self, fmin=0, fmax=numpy.inf) -> DataArray:
        """
        Peak frequency of the spectrum, i.e. frequency at which the spectrum
        obtains its maximum.

        :param fmin: minimum frequency
        :param fmax: maximum frequency
        :return: peak frequency
        """
        return self.peak_frequency(fmin, fmax) * numpy.pi * 2

    def peak_period(self, fmin=0, fmax=numpy.inf) -> DataArray:
        """
        Peak period of the spectrum, i.e. period at which the spectrum
        obtains its maximum.

        :param fmin: minimum frequency
        :param fmax: maximum frequency
        :return: peak period
        """
        peak_period = 1 / self.peak_frequency(fmin, fmax)
        peak_period.name = "peak period"
        peak_period = peak_period.drop("frequency")
        return peak_period

    def peak_direction(self, fmin=0, fmax=numpy.inf) -> DataArray:
        index = self.peak_index(fmin, fmax)
        return self._mean_direction(
            self.a1.isel(**{NAME_F: index}), self.b1.isel(**{NAME_F: index})
        )

    def peak_directional_spread(self, fmin=0, fmax=numpy.inf) -> DataArray:
        index = self.peak_index(fmin, fmax)
        a1 = self.a1.isel(**{NAME_F: index})
        b1 = self.b1.isel(**{NAME_F: index})
        return self._spread(a1, b1)

    @staticmethod
    def _mean_direction(a1: _T, b1: _T) -> _T:
        return numpy.arctan2(b1, a1) * 180 / numpy.pi

    @staticmethod
    def _spread(a1: _T, b1: _T) -> _T:
        return numpy.sqrt(2 - 2 * numpy.sqrt(a1**2 + b1**2)) * 180 / numpy.pi

    @property
    def mean_direction_per_frequency(self) -> DataArray:
        return self._mean_direction(self.a1, self.b1)

    @property
    def mean_spread_per_frequency(self) -> DataArray:
        return self._spread(self.a1, self.b1)

    def _spectral_weighted(self, property: DataArray, fmin=0, fmax=numpy.inf):
        range = {NAME_F: self._range(fmin, fmax)}

        property = property.fillna(0)
        return numpy.trapz(
            property.isel(**range) * self.e.isel(**range), self.frequency[range]
        ) / self.m0(fmin, fmax)

    def mean_direction(self, fmin=0, fmax=numpy.inf):
        return self._mean_direction(self.mean_a1(fmin, fmax), self.mean_b1(fmin, fmax))

    def mean_directional_spread(self, fmin=0, fmax=numpy.inf):
        return self._spread(self.mean_a1(fmin, fmax), self.mean_b1(fmin, fmax))

    def mean_a1(self, fmin=0, fmax=numpy.inf):
        return self._spectral_weighted(self.a1, fmin, fmax)

    def mean_b1(self, fmin=0, fmax=numpy.inf):
        return self._spectral_weighted(self.b1, fmin, fmax)

    def mean_a2(self, fmin=0, fmax=numpy.inf):
        return self._spectral_weighted(self.a2, fmin, fmax)

    def mean_b2(self, fmin=0, fmax=numpy.inf):
        return self._spectral_weighted(self.b2, fmin, fmax)

    @property
    def depth(self) -> DataArray:
        depth = self.dataset[NAME_DEPTH]
        return where(depth.isnull(), numpy.inf, depth)

    @property
    def group_velocity(self) -> DataArray:
        depth = self.depth.expand_dims(dim=NAME_F, axis=-1).values
        k = self.wavenumber.values
        depth = depth * numpy.ones(k.shape)

        # Construct the output coordinates and dimension of the data array
        return_dimensions = (*self.dims_space_time, NAME_F)
        coords = {}
        for dim in return_dimensions:
            coords[dim] = self.dataset[dim].values

        return DataArray(
            data=intrinsic_group_velocity(k, depth),
            dims=return_dimensions,
            coords=coords,
        )

    @property
    def wavenumber(self) -> DataArray:
        """
        Determine the wavenumbers for the frequencies in the spectrum. Note that since the dispersion relation depends
        on depth the returned wavenumber array has the dimensions associated with the depth array by the frequency
        dimension.

        :return: wavenumbers
        """

        # For numba (used in the dispersion relation) we need raw numpy arrays of the correct dimension
        depth = self.depth.expand_dims(dim=NAME_F, axis=-1).values
        radian_frequency = self.radian_frequency.expand_dims(dim=self.depth.dims).values

        # Broadcasting does not work inside the numba implementaiton, we explicitly need to construct arrays of the
        # correct input dimension.
        depth_shape = depth.shape
        radian_frequency_shape = radian_frequency.shape

        depth = depth * numpy.ones(radian_frequency_shape)
        radian_frequency = numpy.ones(depth_shape) * radian_frequency

        # Construct the output coordinates and dimension of the data array
        return_dimensions = (*self.dims_space_time, NAME_F)
        coords = {}
        for dim in return_dimensions:
            coords[dim] = self.dataset[dim].values

        return DataArray(
            data=inverse_intrinsic_dispersion_relation(radian_frequency, depth),
            dims=return_dimensions,
            coords=coords,
        )

    @property
    def wavelength(self) -> DataArray:
        return 2 * numpy.pi / self.wavenumber

    @property
    def peak_wavenumber(self) -> DataArray:
        index = self.peak_index()
        # Construct the output coordinates and dimension of the data array
        coords = {}
        for dim in self.dims_space_time:
            coords[dim] = self.dataset[dim].values

        return DataArray(
            data=inverse_intrinsic_dispersion_relation(
                self.radian_frequency[index].values, self.depth.values
            ),
            dims=self.dims_space_time,
            coords=coords,
        )

    def bulk_variables(self) -> Dataset:
        dataset = Dataset()
        dataset["significant_waveheight"] = self.significant_waveheight
        dataset["mean_period"] = self.mean_period
        dataset["peak_period"] = self.peak_period()
        dataset["peak_direction"] = self.peak_direction()
        dataset["peak_directional_spread"] = self.peak_directional_spread()
        dataset["mean_direction"] = self.mean_direction()
        dataset["mean_directional_spread"] = self.mean_directional_spread()
        dataset["peak_frequency"] = self.peak_frequency()
        dataset["latitude"] = self.latitude
        dataset["longitude"] = self.longitude
        dataset["timestamp"] = self.time
        return dataset

    @property
    def significant_waveheight(self) -> DataArray:
        return self.hm0()

    @property
    def mean_period(self) -> DataArray:
        return self.tm01()

    @property
    def zero_crossing_period(self) -> DataArray:
        return self.tm02()

    def interpolate(self: _T, coordinates) -> _T:
        return self.__class__(interpolate_dataset_grid(coordinates, self.dataset))

    def extrapolate_tail(self, end_frequency, power=None) -> "FrequencySpectrum":
        """
        Extrapolate the tail using the given power
        :param end_frequency: frequency to extrapolate to
        :param power: power to use. If None, a best fit -4 or -5 tail is used.
        :return:
        """
        e = self.e
        a1 = self.a1
        b1 = self.b1
        a2 = self.a2
        b2 = self.b2

        frequency = self.frequency.values
        frequency_delta = frequency[-1] - frequency[-2]
        n = int((end_frequency - frequency[-1]) / frequency_delta) + 1

        fstart = frequency[-1] + frequency_delta
        fend = frequency[-1] + n * frequency_delta
        tail_frequency = numpy.linspace(fstart, fend, n, endpoint=True)
        tail_frequency = DataArray(
            data=tail_frequency, coords={"frequency": tail_frequency}, dims="frequency"
        )
        variance_density = concat(
            (e, e.isel(frequency=-1) * zeros_like(tail_frequency)), dim="frequency"
        )  # (tail_frequency / frequency[-1]) ** power
        a1 = concat(
            (a1, a1.isel(frequency=-1) * ones_like(tail_frequency)), dim="frequency"
        )
        b1 = concat(
            (b1, b1.isel(frequency=-1) * ones_like(tail_frequency)), dim="frequency"
        )
        a2 = concat(
            (a2, a2.isel(frequency=-1) * ones_like(tail_frequency)), dim="frequency"
        )
        b2 = concat(
            (b2, b2.isel(frequency=-1) * ones_like(tail_frequency)), dim="frequency"
        )

        # Find the last nonzero frequency
        variance_density = DataArray(
            data=fill_zeros_or_nan_in_tail(
                variance_density.values, variance_density.frequency.values, power
            ),
            dims=a1.dims,
            coords=a1.coords,
        )

        dataset = Dataset(
            {
                "variance_density": variance_density,
                "a1": a1,
                "b1": b1,
                "a2": a2,
                "b2": b2,
            }
        )

        for name in self.dataset:
            if name in SPECTRAL_VARS:
                continue
            else:
                dataset = dataset.assign({name: self.dataset[name]})

        return FrequencySpectrum(dataset)

    def bandpass(self: _T, fmin=0, fmax=numpy.inf) -> _T:

        dataset = Dataset()

        for name in self.dataset:
            if name in SPECTRAL_VARS:
                data = self.dataset[name].where(
                    (self.frequency >= fmin) & (self.frequency < fmax), drop=True
                )
                dataset = dataset.assign({name: data})
            else:
                dataset = dataset.assign({name: self.dataset[name]})
        cls = type(self)
        return cls(dataset)

    def interpolate_frequency(
        self: _T, new_frequencies, extrapolate_tail=True, extrapolate_power=5
    ) -> _T:
        return self.__class__(
            interpolate_dataset_along_axis(
                new_frequencies, self.dataset, coordinate_name="frequency"
            )
        )

    def _range(self, fmin=0.0, fmax=numpy.inf) -> numpy.ndarray:
        return (self.dataset[NAME_F].values >= fmin) & (
            self.dataset[NAME_F].values < fmax
        )

    def save_as_netcdf(self, path):
        self.dataset.to_netcdf(path)

    def multiply(
        self, array: numpy.ndarray, dimensions: List[str] = None, inplace=False
    ) -> _T:
        """
        Multiply the variance density with the given numpy array. Broadcasting is performed automatically if dimensions
        are provided. If no dimensions are provided the array needs to have the exact same shape as the variance
        density array.

        :param array: Array to multiply with variance density
        :param dimension: Dimensions of the array
        :return: self
        """
        if inplace:
            output = self
        else:
            output = self.copy()

        coords = {}
        shape = array.shape
        if dimensions is None:
            if shape != self.shape():
                raise ValueError(
                    "If no dimensions are provided the array must have the exact same shape as the"
                    "variance density array."
                )

            output.dataset[NAME_E] = self.dataset[NAME_E] * array
            return output

        if len(shape) != len(dimensions):
            raise ValueError(
                "The dimensions of the input array must match the number of dimension labels"
            )

        for length, dimension in zip(shape, dimensions):
            if dimension not in self.dims:
                raise ValueError(
                    f"Dimension {dimension} not a valid dimension of the spectral object."
                )
            coords[dimension] = self.dataset[dimension].values

            if len(self.dataset[dimension].values) != length:
                raise ValueError(
                    f"Array length along the dimension {dimension} does not match the length of the"
                    f"coordinate of the same name in the spctral object."
                )

        data = DataArray(data=array, coords=coords, dims=dimensions)
        output.dataset[NAME_E] = self.dataset[NAME_E] * data
        return output


class FrequencyDirectionSpectrum(WaveSpectrum):
    def __init__(self, dataset: Dataset):
        super(FrequencyDirectionSpectrum, self).__init__(dataset)
        for name in NAMES_2D:
            if name not in dataset and name not in dataset.coords:
                raise ValueError(
                    f"Required variable/coordinate {name} is"
                    f" not specified in the dataset"
                )

    def __len__(self):
        return int(numpy.prod(self.spectral_values.shape[:-2]))

    @property
    def direction_step(self) -> DataArray:
        difference = wrapped_difference(
            numpy.diff(self.direction.values, append=self.direction[0]), period=360
        )
        return DataArray(
            data=difference, coords={NAME_D: self.direction.values}, dims=[NAME_D]
        )

    @property
    def radian_direction(self) -> DataArray:
        data_array = self.dataset[NAME_D] * numpy.pi / 180
        data_array.name = "radian_direction"
        return data_array

    def _directionally_integrate(self, data_array: DataArray) -> DataArray:
        return (data_array * self.direction_step).sum(NAME_D, skipna=True)

    @property
    def e(self) -> DataArray:
        return self._directionally_integrate(self.dataset[NAME_E])

    @property
    def a1(self) -> DataArray:
        return (
            self._directionally_integrate(
                self.dataset[NAME_E] * numpy.cos(self.radian_direction)
            )
            / self.e
        )

    @property
    def b1(self) -> DataArray:
        return (
            self._directionally_integrate(
                self.dataset[NAME_E] * numpy.sin(self.radian_direction)
            )
            / self.e
        )

    @property
    def a2(self) -> DataArray:
        return (
            self._directionally_integrate(
                self.dataset[NAME_E] * numpy.cos(2 * self.radian_direction)
            )
            / self.e
        )

    @property
    def b2(self) -> DataArray:
        return (
            self._directionally_integrate(
                self.dataset[NAME_E] * numpy.sin(2 * self.radian_direction)
            )
            / self.e
        )

    @property
    def direction(self) -> DataArray:
        return self.dataset[NAME_D]

    def as_frequency_spectrum(self) -> "FrequencySpectrum":
        dataset = {
            "a1": self.a1,
            "b1": self.b1,
            "a2": self.a2,
            "b2": self.b2,
            "variance_density": self.e,
        }
        for name in self.dataset:
            if name not in SPECTRAL_VARS:
                dataset[name] = self.dataset[name]

        return FrequencySpectrum(Dataset(dataset))

    def spectrum_1d(self) -> "FrequencySpectrum":
        """
        Will be depricated
        :return:
        """
        warn(
            'spectrum_1d method will be removed, use "as_frequency_spectrum" instead',
            DeprecationWarning,
            stacklevel=2,
        )
        return self.as_frequency_spectrum()

    def differentiate(self, coordinate=None, **kwargs) -> "FrequencyDirectionSpectrum":

        if coordinate is None:
            coordinate = "time"

        if coordinate not in self.dataset:
            raise ValueError(f"Coordinate {coordinate} does not exist in the dataset")

        data = {
            NAME_E: (
                self.dims,
                self.variance_density.differentiate(
                    coordinate, datetime_unit="s", **kwargs
                ).values,
            )
        }
        for x in self.dataset:
            if x in SPECTRAL_VARS:
                continue
            data[x] = (self.dims_space_time, self.dataset[x].values)

        return FrequencyDirectionSpectrum(Dataset(data_vars=data, coords=self.coords()))

    @property
    def number_of_directions(self) -> int:
        return len(self.direction)


class FrequencySpectrum(WaveSpectrum):
    def __init__(self, dataset: Dataset):
        super(FrequencySpectrum, self).__init__(dataset)
        for name in NAMES_1D:
            if name not in dataset and name not in dataset.coords:
                raise ValueError(
                    f"Required variable/coordinate {name} is"
                    f" not specified in the dataset"
                )

    def __len__(self):
        return int(numpy.prod(self.spectral_values.shape[:-1]))

    def interpolate(self: "FrequencySpectrum", coordinates) -> "FrequencySpectrum":
        """

        :param coordinates:
        :return:
        """
        _dataset = Dataset()
        _moments = [NAME_a1, NAME_b1, NAME_a2, NAME_b2]

        # For physical reasons it is better to interpolate the scaled moments - as opposed to the normalized moments.
        # For the dataset we interpolate we set a1: to A1 etc. Afterwards we scale the output back to the normalized
        # state.
        for name in self.dataset:
            _name = str(name)
            if _name in _moments:
                _dataset = _dataset.assign({_name: getattr(self, _name) * self.e})
            else:
                _dataset = _dataset.assign({_name: self.dataset[_name]})

        interpolated_data = interpolate_dataset_grid(coordinates, _dataset)
        for name in _moments:
            interpolated_data[name] = (
                interpolated_data[name] / interpolated_data[NAME_E]
            )

        return FrequencySpectrum(interpolated_data)

    def as_frequency_direction_spectrum(
        self,
        number_of_directions,
        method: Estimators = "mem2",
        solution_method="scipy",
    ) -> "FrequencyDirectionSpectrum":

        direction = numpy.linspace(0, 360, number_of_directions, endpoint=False)

        output_array = (
            estimate_directional_distribution(
                self.a1.values,
                self.b1.values,
                self.a2.values,
                self.b2.values,
                direction,
                method=method,
                solution_method=solution_method,
            )
            * self.e.values[..., None]
        )

        dims = self.dims_space_time + [NAME_F, NAME_D]
        coords = {x: self.dataset[x].values for x in self.dims}
        coords[NAME_D] = direction

        data = {NAME_E: (dims, output_array)}
        for x in self.dataset:
            if x in SPECTRAL_VARS:
                continue
            data[x] = (self.dims_space_time, self.dataset[x].values)

        return FrequencyDirectionSpectrum(Dataset(data_vars=data, coords=coords))


def create_1d_spectrum(
    frequency: numpy.ndarray,
    variance_density: numpy.ndarray,
    time: Union[numpy.ndarray, float],
    latitude: Union[numpy.ndarray, float],
    longitude: Union[numpy.ndarray, float],
    a1: numpy.ndarray = None,
    b1: numpy.ndarray = None,
    a2: numpy.ndarray = None,
    b2: numpy.ndarray = None,
    depth: Union[numpy.ndarray, float] = numpy.inf,
    dims=(NAME_T, NAME_F),
) -> FrequencySpectrum:
    if a1 is None:
        a1 = numpy.nan + numpy.ones_like(variance_density)
    if b1 is None:
        b1 = numpy.nan + numpy.ones_like(variance_density)
    if a2 is None:
        a2 = numpy.nan + numpy.ones_like(variance_density)
    if b2 is None:
        b2 = numpy.nan + numpy.ones_like(variance_density)

    variables = {
        NAME_T: numpy.atleast_1d(to_datetime64(time)),
        NAME_LAT: numpy.atleast_1d(latitude),
        NAME_LON: numpy.atleast_1d(longitude),
        NAME_DEPTH: numpy.atleast_1d(depth),
        NAME_F: frequency,
        NAME_E: variance_density,
        NAME_a1: a1,
        NAME_b1: b1,
        NAME_a2: a2,
        NAME_b2: b2,
    }

    return FrequencySpectrum(create_spectrum_dataset(dims, variables))


def create_2d_spectrum(
    frequency: numpy.ndarray,
    direction: numpy.ndarray,
    variance_density: numpy.ndarray,
    time,
    latitude: Union[numpy.ndarray, float],
    longitude: Union[numpy.ndarray, float],
    dims=(NAME_T, NAME_F, NAME_D),
    depth: Union[numpy.ndarray, float] = numpy.inf,
) -> FrequencyDirectionSpectrum:
    """
    :param frequency:
    :param direction:
    :param variance_density:
    :param time:
    :param latitude:
    :param longitude:
    :param dims:
    :param depth:
    :return:
    """

    variables = {
        NAME_T: numpy.atleast_1d(to_datetime64(time)),
        NAME_LAT: numpy.atleast_1d(latitude),
        NAME_LON: numpy.atleast_1d(longitude),
        NAME_DEPTH: numpy.atleast_1d(depth),
        NAME_F: frequency,
        NAME_D: direction,
        NAME_E: variance_density,
    }
    return FrequencyDirectionSpectrum(create_spectrum_dataset(dims, variables))


def create_spectrum_dataset(dims, variables) -> Dataset:
    independent_variables = []
    for dim in dims:
        if dim in variables:
            independent_variables.append(dim)

    dependent_variables = [x for x in variables if x not in independent_variables]

    spectral_coords = {k: variables[k] for k in independent_variables}
    spatial_coords = {
        k: variables[k] for k in independent_variables if k not in SPECTRAL_DIMS
    }

    dataset = Dataset()
    for variable in dependent_variables:
        if variable in SPECTRAL_VARS:
            coords = spectral_coords
        else:
            coords = spatial_coords
        dims = [k for k in coords]

        if dims:
            dataset = dataset.assign(
                {
                    variable: DataArray(
                        data=variables[variable], dims=dims, coords=coords
                    )
                }
            )

        else:
            if len(variables[variable]) == 1:
                # If no coordinate is known, and the variable has length 1, we add it as a scalar.
                data = variables[variable][0]
            else:
                # otherwise we add without coordinate/dimension.
                data = DataArray(data=variables[variable])

            dataset = dataset.assign({variable: data})
    return dataset


def load_spectrum_from_netcdf(
    filename_or_obj,
) -> Union[FrequencySpectrum, FrequencyDirectionSpectrum]:
    """
    Load a spectrum from netcdf file
    :param filename_or_obj:
    :return:
    """
    dataset = open_dataset(filename_or_obj=filename_or_obj)
    if NAME_D in dataset.coords:
        return FrequencyDirectionSpectrum(dataset=dataset)
    else:
        return FrequencySpectrum(dataset=dataset)



# Logging
from roguewave.log import logger
from roguewave.tools.time import to_datetime_utc, to_datetime64

# model time
from roguewave.modeldata import (
    TimeSliceLead,
    timebase_forecast,
    TimeSliceForecast,
    TimeSliceEvaluation,
    TimeSliceAnalysis,
    model_timebase,
    model_valid_time,
    available_models,
    list_available_variables,
    open_remote_dataset,
    extract_from_remote_dataset,
    generate_uris,
)


# Interpolate
from roguewave.interpolate import (
    interpolate_dataset,
    tracks_as_dataset,
    TrackSet,
    Track,
    Cluster,
)

# Wave spectrum
from roguewave.wavespectra import (
    FrequencySpectrum,
    FrequencyDirectionSpectrum,
    create_1d_spectrum,
    create_2d_spectrum,
    concatenate_spectra,
    load_spectrum_from_netcdf,
    WaveSpectrum,
    integrate_spectral_data,
    SPECTRAL_DIMS,
    NAME_F,
    NAME_D,
)

# IO
from .io.io import save, load

# Restart Files:
from roguewave.wavewatch3 import (
    RestartFile,
    open_restart_file,
    write_partial_restart_file,
    write_restart_file,
    reassemble_restart_file_from_parts,
    unpack_ww3_data,
    read_model_definition,
)

# Partitioning
from roguewave.wavespectra.partitioning import (
    get_bulk_partitions_from_spectral_partitions,
    get_spectral_partitions_from_2dspectra,
    get_bulk_partitions_from_observations,
    get_spectral_partitions_from_observations,
)

# Colocation
from roguewave.colocate import (
    colocate_model_spotter,
    colocated_tracks_as_dataset,
    colocate_model_spotter_spectra,
)
from roguewave.spotterapi import get_spotter_data, get_spectrum
from roguewave.spotter import spotter_frequency_response_correction
from roguewave.observations import get_satellite_data, get_satellite_available
from roguewave.wavespectra.estimators import estimate_directional_spectrum_from_moments



from json import dumps

_format = {
    "sampling_interval_seconds": 3600,
    "sampling_interval_gps": 0.4,
    "nfft": 256,
    "time_column": "t0_GPS_Epoch_Time(s)",
    "pattern": "????_SPC.csv",
    "columns": [
        {
            "name": "type",
            "dtype": "str",
        },
        {
            "name": "millis",
            "dtype": "float64",
        },
        {
            "name": "t0_GPS_Epoch_Time(s)",
            "dtype": "float64",
        },
        {
            "name": "tN_GPS_Epoch_Time(s)",
            "dtype": "float64",
        },
        {
            "name": "ens_count",
            "dtype": "int64",
        },
    ],
}

_spectral_column_names = {
    "default": [
        "Sxx_re",
        "Syy_re",
        "Szz_re",
        "Sxy_re",
        "Szx_re",
        "Szy_re",
        "Sxx_im",
        "Syy_im",
        "Szz_im",
        "Sxy_im",
        "Szx_im",
        "Szy_im",
    ]
}


def spectral_column_names(groupedby="frequency"):
    names = _spectral_column_names["default"]

    columns = []
    for index in range(0, _format["nfft"] // 2):
        columns += [name + f"_{index}" for name in names]
    return columns


def create_spectral_format():
    def _create(name):
        return {
            "name": name,
            "dtype": "float64",
        }

    column_names = spectral_column_names("flat")
    _format["columns"] += [_create(name) for name in column_names]
    return _format


if __name__ == "__main__":
    format = create_spectral_format()
    with open("SPC.json", "wt") as file:
        str = dumps(format, indent=4)
        file.write(str)



"""
Contents: Partitioning

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
======================

Partitioning routines that can be used to partition the spectrum.

Classes:

- `SeaSwellData`, return type that contains the partitioned data.
- `Partition`, data class that describes a partition

Functions:

- `neighbours`,

How To Use This Module
======================
(See the individual functions for details.)

1. Import it: ``import partitioning`` or ``from partitioning import ...``.
2.
"""

from roguewave.wavespectra import FrequencyDirectionSpectrum
import numpy
import numba.typed
from typing import Dict, List, Tuple
import numba
from scipy.ndimage import maximum_filter, generate_binary_structure

default_partition_config = {
    'minimumEnergyFraction': 0.001,
    'minimumDensityRatio': 0.
}

NOT_ASSIGNED = -1


@numba.njit(cache=True)
def floodfill(spectral_density: numpy.ndarray, min_val=0.0) -> Tuple[
    numpy.ndarray, Dict[int, List[int]], Dict[int, Tuple[int, int]]]:
    """
    Flood fill algorithm. We try to find the region that belongs to a peak 
    according to inverse watershed.The implementation has been optimized for 
    use with numba.

    :param spectral_density: 2d-ndarray, first dimension frequency, second direction
    :param min_val: minimum value we consider in the algorithm.

    :return:

    We start at a local peak indicated by [peak_frequency_index]
    and [peak_direction_index], label that peak with the given [label]
    """

    number_of_directions = spectral_density.shape[1]
    number_of_frequencies = spectral_density.shape[0]

    proximate_partitions = {}  # type: Dict[int,List[int]]

    partition_label = numpy.zeros(
        (number_of_frequencies, number_of_directions),
        dtype='int32') + NOT_ASSIGNED

    current_label = 0
    assigned = False
    for start_frequency_index in range(0, number_of_frequencies):
        for start_direction_index in range(0, number_of_directions):
            if spectral_density[
                start_frequency_index, start_direction_index] == min_val:
                assigned = True
                partition_label[
                    start_frequency_index, start_direction_index] = current_label

    if assigned:
        proximate_partitions[current_label] = numpy.array([0], dtype='int64')

    #
    # We loop over all cells in the spectrum...
    peak_indices = {}
    # delta = numpy.zeros( 8 )
    ii = numpy.zeros(1000, dtype='int64')
    jj = numpy.zeros(1000, dtype='int64')

    neighbour_direction_offset = numpy.array(
        (0, 0, - 1, - 1, - 1, + 1, + 1, + 1), dtype='int64')
    neighbour_frequency_offset = numpy.array(
        (- 1, + 1, - 1, 0, + 1, - 1, 0, + 1), dtype='int64')
    neighbour_frequency_indices = numpy.zeros(8, dtype='int64')
    neighbour_direction_indices = numpy.zeros(8, dtype='int64')

    for start_frequency_index in range(0, number_of_frequencies):
        for start_direction_index in range(0, number_of_directions):

            # if the cell is already assigned to a peak during a previous ascent
            # from another staring point we can skip the current cell.
            if partition_label[
                start_frequency_index, start_direction_index] > NOT_ASSIGNED:
                continue

            # Initialize the path. We keep track of the visited cells in an
            # ascent by denoting their indices in a list.
            # ii = [start_frequency_index]
            # jj = [start_direction_index]
            ii[0] = start_frequency_index
            jj[0] = start_direction_index
            number_of_points_in_path = 1

            # In addition- as we climb the hill we denote any partitions we
            # happen to neighbour
            proximate_partitions_work = []

            while True:
                # get the last cell in the path
                direction_index = jj[number_of_points_in_path - 1]  # jj[-1]
                frequency_index = ii[number_of_points_in_path - 1]  # ii[-1]

                if (partition_label[
                    frequency_index, direction_index] > NOT_ASSIGNED):
                    label = partition_label[frequency_index, direction_index]
                    proximate_partitions[label] = \
                        numpy.append(proximate_partitions[label],
                                     proximate_partitions_work)

                    for index in range(0, number_of_points_in_path):
                        partition_label[ii[index], jj[index]] = label
                    break

                neighbour_frequency_indices[
                :] = neighbour_frequency_offset + frequency_index
                neighbour_direction_indices[:] = (
                                                         neighbour_direction_offset + direction_index) % number_of_directions

                # Get the value of the energy density at the current cell
                node_value = spectral_density[frequency_index, direction_index]

                # Here we calculate the differences in energy density from the
                # current cell to each of it neighbours.
                # delta[:] = -numpy.inf
                all_smaller_than_zero = True
                index_max_delta = -1
                max_delta = -numpy.inf
                for index, ifreq, idir in zip(
                        range(0, len(neighbour_frequency_indices)),
                        neighbour_frequency_indices,
                        neighbour_direction_indices):

                    if ifreq < 0 or ifreq >= number_of_frequencies:
                        continue

                    # Calculate the delta for each neighbour
                    delta = (spectral_density[ifreq, idir] - node_value)
                    if delta > max_delta:
                        index_max_delta = index
                        max_delta = delta

                    if delta > 0:
                        all_smaller_than_zero = False

                    if partition_label[ifreq, idir] > 0:
                        proximate_partitions_work.append(
                            partition_label[ifreq, idir])

                if all_smaller_than_zero:
                    # this is a peak
                    current_label += 1
                    label = current_label
                    proximate_partitions[label] = numpy.array(
                        proximate_partitions_work, dtype='int64')
                    peak_indices[label] = (frequency_index, direction_index)

                    for index in range(0, number_of_points_in_path):
                        partition_label[ii[index], jj[index]] = label

                    break
                else:
                    # steepest_index = numpy.argmax(delta)
                    ii[number_of_points_in_path] = neighbour_frequency_indices[
                        index_max_delta]
                    jj[number_of_points_in_path] = neighbour_direction_indices[
                        index_max_delta]
                    number_of_points_in_path += 1

                    if number_of_points_in_path == len(ii):
                        # because we preallocate the path arrays we have to
                        # make sure they fit- and otherwise extend if need be
                        _ii = numpy.zeros(2 * len(ii), dtype='int64')
                        _jj = numpy.zeros(2 * len(ii), dtype='int64')
                        _ii[0:number_of_points_in_path] = ii
                        _jj[0:number_of_points_in_path] = jj
                        ii = _ii
                        jj = _jj

                    # ii.append(neighbour_frequency_indices[steepest_index])
                    # jj.append(neighbour_direction_indices[steepest_index])
                    continue
        #

    for label in proximate_partitions:
        proximate_partitions[label] = numpy.unique(proximate_partitions[label])
        mask = proximate_partitions[label] == label
        proximate_partitions[label] = proximate_partitions[label][~mask]

    for label_source, item in proximate_partitions.items():
        for label_target in item:
            if label_source == label_target:
                continue

            if label_source not in proximate_partitions[label_target]:
                proximate_partitions[label_target] = numpy.append(
                    proximate_partitions[label_target], label_source)

    return partition_label, proximate_partitions, peak_indices


@numba.njit(cache=True)
def find_label_closest_partition(label,
                                 proximate_labels: List[int],
                                 number_of_directions: int,
                                 peak_indices: Dict[int, Tuple[int, int]]):
    """
    Find the partition whose peak is "closest" to the partition under consideration
    Only proximate partitions are considered. Basically: find the partition whose
    maximum is closest to the peak of the current partition.

    :param label:
    :param proximate_labels:
    :param number_of_directions:
    :param peak_indices:

    :return:
    """

    if not proximate_labels:
        return None

    # Initialize the numpy arrays for distance and indices.
    distance = numpy.zeros((len(proximate_labels),))
    indices = numpy.zeros((len(proximate_labels),), dtype='int64')

    ifreq, idir = peak_indices[label]
    for ii, local_label in enumerate(proximate_labels):
        delta_freq = peak_indices[local_label][0] - ifreq
        delta_dir = (peak_indices[local_label][
                         1] - idir - number_of_directions // 2) % number_of_directions + number_of_directions // 2
        distance[ii] = numpy.sqrt(delta_freq ** 2 + delta_dir ** 2)
        indices[ii] = local_label

    return indices[numpy.argmin(distance)]


def filter_for_low_energy(partitions: Dict[int, FrequencyDirectionSpectrum],
                          proximity: Dict[int, List[int]],
                          peak_indices: Dict[int, Tuple[int, int]],
                          total_energy,
                          config):
    """
    Remove partitions that contain only a small fraction of the total energy
    in the spectrum.

    :param partitions: Dictionary of spectra, keys are the labels identifying
                       the partition under consideration

    :param proximity: Dictionary that denotes for each partition identified by
                      its label (dict key) a list of partitions that are direct
                      neighbours

    :param peak_indices: Dictionary that for each partition identified by its
                         label (dict key) the frequency and direction index of
                         the peak value

    :param total_energy: Total energy (variance) in the wave spectrum

    :param config: Configuration dictionary

    :return: None (function has side effects)
    """

    # If there is only one partition, there is no need to filter, return
    if len(partitions) == 1:
        return

    # loop over all the partition labels
    for label in list(partitions.keys()):

        # This looks weird- but is needed. The dictionary is changed and a label
        # may no longer exist.
        if label not in partitions:
            continue

        # get the partition associated with the label
        partition = partitions[label]

        # Check if the partition contains sufficient energy
        if (partition.m0() < config['minimumEnergyFraction'] * total_energy):
            #
            # If not, Merge with the closest partition. First find what is the
            # closest neighbour.
            if partition.hm0() == 0:
                closest_label = proximity[label][0]
            else:
                # Note we have to pass a numba typed list to avoid the deprecation
                # warning on general python lists.

                if not len(proximity[label]) == 0:
                    closest_label = find_label_closest_partition(
                        label, numba.typed.List(proximity[label]),
                        len(partition.direction), peak_indices)
                else:
                    closest_label = None

            # There is an edge case where there are no neighbours to merge with
            # hence the check
            if closest_label is not None:
                # if a closest partition is found- lets merge with that
                # partition.
                merge_partitions(label, closest_label, partitions,
                                 proximity)


def merge_partitions(
        source_label,
        target_label,
        partitions: Dict[int, FrequencyDirectionSpectrum],
        proximity: Dict[int, List[int]]) -> None:
    """
    Merge one partition into another partition. (**Has side-effects**)
    :param source_index: source index of the partition to merge in the list
    :param target_index: target index of the partition that is merged into
    :param partitions: the list of partitions
    :param partition_label_array: 2D array of labels indicating to what partition
    the cel belongs

    :return: None
    """

    # Add the source to the target spectrum.
    partitions[target_label] = partitions[target_label] + partitions[
        source_label]

    # Remove the source from the partition dict
    partitions.pop(source_label)

    # The merge is now complete, but we need to update the adjacency list that
    # denotes which partitions border one-another since anything that bordered
    # the source partition will now border the target partition. Furter, we need
    # to remove references to the source partition in the adjacency list.

    # loop over all partitions bordering the source partition
    for proximate_labels_to_source_label in proximity[source_label]:

        # for the neighbour, get its list of partitions that border it.
        proximate_to_update = proximity[proximate_labels_to_source_label]

        # find the index in this list that refers to the source
        index = proximate_to_update.index(source_label)

        # If this is the target partition, merely remove the reference to the
        # source partition from the list
        if proximate_labels_to_source_label == target_label:
            proximity[target_label].pop(index)
            continue

        #
        # Make sure the partition listed as bordering the target partition.
        if proximate_labels_to_source_label not in proximity[target_label]:
            proximity[target_label].append(proximate_labels_to_source_label)

        # If the target already is listed as bordering, merely remove the
        # reference of the source from the list
        if target_label in proximate_to_update:
            proximate_to_update.pop(index)
        else:
            # if it does not border, update the reference to the source with
            # a reference to the target.
            proximate_to_update[index] = target_label

    # We can now safely remove the list from the proximity dict.
    proximity.pop(source_label)
    return None


def partition_spectrum(spectrum: FrequencyDirectionSpectrum, config=None) -> \
        Tuple[Dict[int, FrequencyDirectionSpectrum], Dict[
            int, List[int]]]:
    """
    Create partitioned spectra from a given 2D wavespectrum
    :param spectrum: 2D wavespectrum
    :param config: configuration parameters - see "default_partition_config"
    default variable.

    :return:
    """

    # Update config if provided
    config = default_partition_config | (config if config else {})

    # Make sure there are no NaN (undifined) values
    density = spectrum.spectral_values.values
    density[numpy.isnan(density)] = 0.0

    # Get partition descriptions using floodfill. This returns an array that
    # labels for each cell in the spectrum to which partition it belongs, an
    # ordered list with the label as index containing which partitions are
    # adjacent to the current partition and the labels.
    partition_label, proximate, peak_indices = floodfill(density)

    # We need to convert back to a list, and we need to do it in a somewhat
    # roundabout way as numba does not support dicts of lists (note, proximate
    # currently is a "numba" dictionary of numpy arrays)
    proximate_partitions = {}
    for label in proximate:
        proximate_partitions[label] = list(proximate[label])

    # Create Partition objects given the label array from the floodfill.
    partitions = {}

    # Create a dict of all draft partitions
    for label, proximity_list in proximate_partitions.items():
        if label == 0:
            continue
        mask = partition_label == label
        partitions[label] = spectrum.extract(mask)
    #
    # # merge low energy partitions
    filter_for_low_energy(partitions, proximate_partitions, peak_indices,
                         spectrum.m0(), config)

    # Return the partitions and the label array.
    return partitions, proximate_partitions


def partition_marker_array(
        partitions: Dict[int, FrequencyDirectionSpectrum]) -> numpy.ndarray:
    a_label = list(partitions.keys())[0]
    marker_array = numpy.zeros(
        partitions[a_label].spectral_values.shape) + numpy.nan
    for label, partition in partitions.items():
        mask = ~partition.spectral_values.mask
        marker_array[mask] = label
    return marker_array


def sum_partitions(
        partitions: Dict[int,FrequencyDirectionSpectrum]) \
        -> FrequencyDirectionSpectrum:
    key = list(partitions.keys())[0]
    sum_spec = partitions[key].copy(deep=True)
    sum_spec['variance_density'] *= 0
    for _, spec in partitions.items():
        sum_spec = spec + sum_spec
    return sum_spec


def is_neighbour(partition_1:FrequencyDirectionSpectrum,
                 partition_2: FrequencyDirectionSpectrum) -> int:
    footprint = generate_binary_structure(partition_2.spectral_values.ndim, 2)

    mask_1 = maximum_filter(partition_1.spectral_values.filled(0),
                            footprint=footprint,
                            mode='wrap')
    mask_2 = partition_2.spectral_values.filled(0)

    return numpy.nansum(mask_1 * mask_2)



from roguewave.wavephysics.fluidproperties import (
    AIR,
    WATER,
    GRAVITATIONAL_ACCELERATION,
)

from roguewave.wavephysics.balance import WindGeneration
from numpy import cos, pi, log, exp, empty, inf
from numba import njit
from roguewave.wavetheory import inverse_intrinsic_dispersion_relation
from typing import TypedDict


class ST4WaveGenerationParameters(TypedDict):
    gravitational_acceleration: float
    charnock_maximum_roughness: float
    charnock_constant: float
    air_density: float
    water_density: float
    vonkarman_constant: float
    wave_age_tuning_parameter: float
    growth_parameter_betamax: float
    elevation: float


class ST4WindInput(WindGeneration):
    name = "st4 generation"

    def __init__(self, parameters: ST4WaveGenerationParameters = None):
        super(ST4WindInput, self).__init__(parameters)
        self._wind_source_term_function = _st4_wind_generation_point

    @staticmethod
    def default_parameters() -> ST4WaveGenerationParameters:
        return ST4WaveGenerationParameters(
            wave_age_tuning_parameter=0.006,
            growth_parameter_betamax=1.52,
            gravitational_acceleration=GRAVITATIONAL_ACCELERATION,
            charnock_maximum_roughness=inf,
            charnock_constant=0.01,
            air_density=AIR.density,
            water_density=WATER.density,
            vonkarman_constant=AIR.vonkarman_constant,
            elevation=10,
        )


# ----------------------------------------------------------------------------------------------------------------------
# ST4 Point wise implementation functions (apply to a single spatial point)
# ----------------------------------------------------------------------------------------------------------------------
@njit(cache=True, fastmath=True)
def _st4_wind_generation_point(
    variance_density,
    wind,
    depth,
    roughness_length,
    st4_spectral_grid,
    st4_parameters,
    wind_source=None,
):
    """
    Implememtation of the st4 wind growth rate term

    :param variance_density:
    :param wind:
    :param depth:
    :param roughness_length:
    :param st4_spectral_grid:
    :param st4_parameters:
    :return:
    """

    # Get the spectral size
    number_of_frequencies, number_of_directions = variance_density.shape

    # Allocate the output variable
    if wind_source is None:
        wind_source = empty((number_of_frequencies, number_of_directions))

    # Unpack variables passed in dictionaries/tuples for ease of reference and some slight speed benifits
    # (avoid dictionary lookup in loops).
    vonkarman_constant = st4_parameters["vonkarman_constant"]
    growth_parameter_betamax = st4_parameters["growth_parameter_betamax"]
    air_density = st4_parameters["air_density"]
    water_density = st4_parameters["water_density"]
    wave_age_tuning_parameter = st4_parameters["wave_age_tuning_parameter"]
    gravitational_acceleration = st4_parameters["gravitational_acceleration"]
    radian_frequency = st4_spectral_grid["radian_frequency"]
    radian_direction = st4_spectral_grid["radian_direction"]
    elevation = st4_parameters["elevation"]

    wind_forcing, wind_direction_degrees, wind_forcing_type = wind

    # Preprocess winds to the correct conventions (U10->friction velocity if need be, wind direction degrees->radians)
    if wind_forcing_type == "u10":
        friction_velocity = (
            wind_forcing * vonkarman_constant / log(elevation / roughness_length)
        )

    elif wind_forcing_type in ["ustar", "friction_velocity"]:
        friction_velocity = wind_forcing

    else:
        raise ValueError("Unknown wind input type")

    wind_direction_radian = wind_direction_degrees * pi / 180

    # precalculate the constant multiplication.
    constant_factor = (
        growth_parameter_betamax / vonkarman_constant**2 * air_density / water_density
    )

    if depth == inf:
        wavenumber = radian_frequency**2 / gravitational_acceleration
    else:
        wavenumber = inverse_intrinsic_dispersion_relation(radian_frequency, depth)

    cosine = cos(radian_direction - wind_direction_radian)

    # Loop over all frequencies/directions
    for frequency_index in range(number_of_frequencies):

        # Since wavenumber and wavespeed only depend on frequency we calculate those outside of the direction loop.
        constant_frequency_factor = constant_factor * radian_frequency[frequency_index]
        relative_speed = (
            friction_velocity
            * wavenumber[frequency_index]
            / radian_frequency[frequency_index]
        )

        for direction_index in range(number_of_directions):
            # If the wave direction has an along wind component we continue.
            if cosine[direction_index] > 0:

                # Calculate the directional factor in the wind input formulation
                W = relative_speed * cosine[direction_index]

                effective_wave_age = log(
                    wavenumber[frequency_index] * roughness_length
                ) + vonkarman_constant / (W + wave_age_tuning_parameter)

                if effective_wave_age > 0:
                    effective_wave_age = 0

                growth_rate = (
                    constant_frequency_factor
                    * exp(effective_wave_age)
                    * effective_wave_age**4
                    * W**2
                )

                wind_source[frequency_index, direction_index] = (
                    growth_rate * variance_density[frequency_index, direction_index]
                )
            else:
                wind_source[frequency_index, direction_index] = 0.0

    return wind_source



from .spectrum import (
    FrequencyDirectionSpectrum,
    FrequencySpectrum,
    create_1d_spectrum,
    create_2d_spectrum,
    WaveSpectrum,
    load_spectrum_from_netcdf,
    NAME_D,
    NAME_F,
    NAME_E,
    NAME_T,
    NAME_LAT,
    NAME_LON,
    NAME_DEPTH,
    NAME_a1,
    NAME_b1,
    NAME_a2,
    NAME_b2,
    NAMES_1D,
    NAMES_2D,
    SPECTRAL_DIMS,
    SPECTRAL_VARS,
    SPACE_TIME_DIMS,
)

from .operations import concatenate_spectra, integrate_spectral_data



from roguewave.colocate.bulk import colocate_model_spotter, colocated_tracks_as_dataset
from roguewave.colocate.spectra import colocate_model_spotter_spectra
from roguewave.colocate.pointdata import colocate_points



from roguewave import FrequencyDirectionSpectrum
from roguewave.wavephysics.fluidproperties import (
    AIR,
    FluidProperties,
    GRAVITATIONAL_ACCELERATION,
)
from roguewave.tools.solvers import fixed_point_iteration
from roguewave.wavephysics.balance import SourceTermBalance
from xarray import DataArray
from numpy import log, inf, sqrt, exp


def drag_coefficient_wu(speed):
    return (0.8 + 0.065 * speed) / 1000


def roughness_wu(speed, elevation, air: FluidProperties = AIR):
    drag = sqrt(drag_coefficient_wu(speed))
    return elevation / exp(air.vonkarman_constant / drag)


def charnock_roughness_length(friction_velocity: DataArray, **kwargs) -> DataArray:
    charnock_constant = kwargs.get("charnock_constant", 0.012)
    gravitational_acceleration = kwargs.get(
        "gravitational_acceleration", GRAVITATIONAL_ACCELERATION
    )

    air_kinematic_viscosity = kwargs.get(
        "air_kinematic_viscosity", AIR.kinematic_viscosity
    )
    viscous_constant = kwargs.get("viscous_constant", 0.0)

    z_visc = viscous_constant * air_kinematic_viscosity / friction_velocity
    z_visc = z_visc.where(friction_velocity > 0, 0.0)

    return (
        charnock_constant * friction_velocity**2 / gravitational_acceleration + z_visc
    )


def charnock_roughness_length_from_u10(speed, **kwargs) -> DataArray:
    air = kwargs.get("air", AIR)
    elevation = kwargs.get("elevation", 10)
    guess = kwargs.get("guess", roughness_wu(speed, elevation, air.vonkarman_constant))

    def _func(roughness):
        friction_velocity = air.vonkarman_constant * speed / log(elevation / roughness)
        return roughness(friction_velocity, **kwargs)

    return fixed_point_iteration(
        _func, guess, bounds=(0, inf), caller="roughness_from_speed", **kwargs
    )


def janssen_roughness_length(
    friction_velocity: DataArray,
    spectrum: FrequencyDirectionSpectrum,
    balance: SourceTermBalance,
    wind_direction: DataArray = None,
):
    if wind_direction is None:
        wind_direction = balance.dissipation.mean_direction_degrees(spectrum)

    return balance.generation.roughness(
        friction_velocity,
        direction=wind_direction,
        spectrum=spectrum,
        wind_speed_input_type="friction_velocity",
    )


def janssen_roughness_length_from_u10(
    friction_velocity: DataArray,
    spectrum: FrequencyDirectionSpectrum,
    balance: SourceTermBalance,
    wind_direction: DataArray = None,
    **kwargs
):
    if wind_direction is None:
        wind_direction = balance.dissipation.mean_direction_degrees(spectrum)

    return balance.generation.roughness(
        friction_velocity,
        direction=wind_direction,
        spectrum=spectrum,
        wind_speed_input_type="friction_velocity",
    )


def drag_coefficient(u10: DataArray, roughness: DataArray, **kwargs) -> DataArray:

    air = kwargs.get("air", AIR)
    elevation = kwargs.get("elevation", 10)

    friction_velocity = u10 * air.vonkarman_constant / log(elevation / roughness)
    return (friction_velocity / u10) ** 2



from xarray import DataArray
from typing import Union

GRAVITATIONAL_ACCELERATION = 9.81
WATER_DENSITY = 1024
WATER_TEMPERATURE_DEGREES_C = 15
AIR_TEMPERATURE_DEGREES_C = 15
AIR_DENSITY = 1.225
KINEMATIC_VISCOSITY_WATER = 1.19e-6
KINEMATIC_VISCOSITY_AIR = 1.48 * 10**-5
VONKARMAN_CONSTANT = 0.4

_input_type = Union[DataArray, float]


class FluidProperties:
    def __init__(
        self,
        density: _input_type,
        temperature: _input_type,
        kinematic_viscosity: _input_type,
        vonkarman_constant: _input_type,
    ):
        self.density = density
        self.temperature = temperature
        self.kinematic_viscosity = kinematic_viscosity
        self.vonkarman_constant = vonkarman_constant


AIR = FluidProperties(
    density=AIR_DENSITY,
    temperature=AIR_TEMPERATURE_DEGREES_C,
    kinematic_viscosity=KINEMATIC_VISCOSITY_AIR,
    vonkarman_constant=VONKARMAN_CONSTANT,
)

WATER = FluidProperties(
    density=WATER_DENSITY,
    temperature=WATER_TEMPERATURE_DEGREES_C,
    kinematic_viscosity=KINEMATIC_VISCOSITY_WATER,
    vonkarman_constant=VONKARMAN_CONSTANT,
)



from .read_csv_data import (
    read_spectra,
    read_displacement,
    read_gps,
    read_raw_spectra,
    read_baro,
    read_raindb,
    read_baro_raw,
    read_sst,
    read_gmn,
    read_location,
    read_and_concatenate_spotter_csv,
)

from .analysis import (
    spectra_from_raw_gps,
    spectra_from_displacement,
    displacement_from_gps_doppler_velocities,
    displacement_from_gps_positions,
    spotter_frequency_response_correction,
)



from dataclasses import dataclass
from typing import Union, Sequence, TypeVar

import numpy
from xarray import DataArray, Dataset

from roguewave.tools.math import wrapped_difference
from roguewave import FrequencyDirectionSpectrum
from roguewave.tools.time import to_datetime64


@dataclass()
class Grid:
    """
    A class that encapsulates all grid data from the model definition. It also
    contains convinience methods to project linear indexed data to a 2d array
    of latitude and longitude (with NaN for missing values), and vice versa
    extract the needed data to construct a linear indexed version compatible
    with WW3.

    Purpose:
    We need this information for reading restart files.
    """

    number_of_spatial_points: int  # number of spatial "sea" points
    frequencies: numpy.ndarray
    directions: numpy.ndarray
    latitude: numpy.ndarray  # 1D array of model latitudes
    longitude: numpy.ndarray  # 1D array of model longitudes
    depth: numpy.ndarray
    mask: numpy.ndarray
    _to_linear_index: numpy.ndarray  # mapping of [ilon,ilat] => linear index
    _to_point_index: numpy.ndarray  # mapping of linear index => [ilon,ilat]

    @property
    def _growth_factor(self):
        return self.frequencies[1] / self.frequencies[0]

    def longitude_index(
        self, linear_index: Union[slice, Sequence, numpy.ndarray]
    ) -> numpy.ndarray:
        """
        :param linear_index: linear 1d index
        :return: longitude 1d index
        """

        return self._to_point_index[0, linear_index]

    def latitude_index(
        self, linear_index: Union[slice, Sequence, numpy.ndarray]
    ) -> numpy.ndarray:
        """
        :param linear_index: linear 1d index
        :return: latitude 1d index
        """
        return self._to_point_index[1, linear_index]

    def index(
        self,
        latitude_index: Union[slice, Sequence, numpy.ndarray],
        longitude_index: Union[slice, Sequence, numpy.ndarray],
        valid_only=False,
    ):
        """
        get the linear index of the array
        :param latitude_index:
        :param longitude_index:
        :return:
        """
        indices = self._to_linear_index[latitude_index, longitude_index]
        if valid_only:
            indices = indices[indices >= 0]
        return indices

    def direction_step(self):
        return wrapped_difference(
            numpy.diff(self.directions, append=self.directions[0]), period=360
        )

    def frequency_step(self):
        delta = numpy.diff(
            self.frequencies,
            prepend=self.frequencies[0] / self._growth_factor,
            append=self.frequencies[-1] * self._growth_factor,
        )
        return (delta[0:-1] + delta[1:]) / 2

    def extract(
        self,
        s: Union[slice, Sequence, numpy.ndarray],
        var: Union[numpy.ndarray, DataArray],
    ) -> DataArray:
        """
        Extract linear indexed data (with indices indicated by the slice) from
        a 2D array (lat,lon)
        :param s: slice instance
        :param var: 2D ndarray
        :return: 1D ndarray of length slice.stop-slice.start
        """
        if isinstance(var, DataArray):
            return var[
                DataArray(self.latitude_index(s), dims="points"),
                DataArray(self.longitude_index(s), dims="points"),
            ]
        else:
            return DataArray(
                data=var[self.latitude_index(s), self.longitude_index(s)], dims="points"
            )

    def project(
        self,
        lon_slice: slice,
        lat_slice: slice,
        var: Union[numpy.ndarray, DataArray],
        except_val=numpy.nan,
    ) -> DataArray:
        """
        Project linear indexed data onto a 2d array (lat ,lon).

        :param lon_slice: slice of ilon indices we want
        :param lat_slice: slice of ilat indices we want
        :param var:
        :param except_val:
        :return:
        """
        ind = self._to_linear_index[lat_slice, lon_slice]
        mask = ind >= 0
        ind = ind[mask]
        ilon = self.longitude_index(ind)
        ilat = self.latitude_index(ind)

        nlon = len(self.longitude[lon_slice])
        nlat = len(self.latitude[lat_slice])

        # this will not work for stepsizes != 1
        out = numpy.zeros((nlat, nlon), dtype=var.dtype) + except_val
        out[ilat, ilon] = var[ind]

        return DataArray(
            data=out,
            coords={
                "latitude": self.latitude[lat_slice],
                "longitude": self.longitude[lon_slice],
            },
            dims=("latitude", "longitude"),
            name="variance",
        )

    def set_linear_data(
        self,
        lon_slice: Union[slice, Sequence, numpy.ndarray],
        lat_slice: Union[slice, Sequence, numpy.ndarray],
        linear_data: Union[numpy.ndarray, DataArray],
        data: Union[numpy.ndarray, DataArray],
    ):
        """
        Update data stored in a linear array with data from a 2d array.
        :param lon_slice: longitude indices
        :param lat_slice: latitude indices
        :param linear_data: linear data array to update
        :param data: 2d Array
        :return:
        """

        ind = self.index(lat_slice, lon_slice, valid_only=True)
        linear_data[ind] = self.extract(ind, data)
        return linear_data


def _unpack_ndarray(data: numpy.ndarray, grid: Grid, exception_value) -> numpy.ndarray:

    shape = data.shape
    if shape[0] != grid.number_of_spatial_points:
        raise ValueError(
            "First dimension of data needs to have the same number of points as the number of sea points"
            f"in the ww3 model. Given: {shape[0]}, expected: {grid.number_of_spatial_points}"
        )

    output = (
        numpy.zeros((len(grid.latitude), len(grid.longitude), *shape[1:]))
        + exception_value
    )
    ind = grid.index(slice(None, None, None), slice(None, None, None), True)

    ilon = grid.longitude_index(ind)
    ilat = grid.latitude_index(ind)
    output[ilat, ilon, ...] = data[:, ...]
    return output


def _unpack_data_array(data: DataArray, grid: Grid, exception_value) -> DataArray:
    dims = []
    output_lat_index = 0
    output_lon_index = 1
    for index, dim in enumerate(data.dims):
        if dim == "linear_index":
            dims += ["latitude", "longitude"]
            output_lat_index = index
            output_lon_index = index + 1
        else:
            dims += [dim]

    coords = {x: data[x] for x in data.dims if x != "linear_index"}
    coords["latitude"] = grid.latitude
    coords["longitude"] = grid.longitude

    output_shape = [len(coords[x]) for x in dims]
    output = numpy.zeros(output_shape) + exception_value

    ind = grid.index(slice(None, None, None), slice(None, None, None), True)
    ilon = grid.longitude_index(ind)
    ilat = grid.latitude_index(ind)

    output_indicer = [slice(None)] * len(output_shape)
    output_indicer[output_lat_index] = ilat
    output_indicer[output_lon_index] = ilon
    output[tuple(output_indicer)] = data.values
    return DataArray(data=output, dims=dims, coords=coords)


_T = TypeVar("_T", DataArray, numpy.ndarray, FrequencyDirectionSpectrum)


def unpack_ww3_data(data: _T, grid: Grid, exception_value=numpy.nan) -> _T:
    """
    Unpack wavewatch3 data stored on a linear grid. Data can scalars (e.g. waveheigths) or spectra. Formats can be numpy
    arrays, dataarrays or spectra object. The output replaces the linear indexed dimension with unpacked two dimensions
    (lat,lon). Locations where data is not given are set to the exception value. To note: if a numpy array is given,
    the linear indexed dimension must be the first dimension.

    :param data: data to unpack
    :param grid: ww3 grid description
    :return: unpacked data.
    """
    if isinstance(data, DataArray):
        return _unpack_data_array(data, grid, exception_value)

    elif isinstance(data, numpy.ndarray):
        return _unpack_ndarray(data, grid, exception_value)

    elif isinstance(data, FrequencyDirectionSpectrum):
        # For the spectral object we need to unpack the variance density, and the depth together.
        variance_density = _unpack_data_array(
            data.variance_density, grid, exception_value
        )
        depth = _unpack_data_array(data.depth, grid, exception_value)

        # Creating the output dataset. Note that "time" is not necessarily a dimension for spectra returned from the
        # wavewatch 3 restart file, but is a required entry. So we may need to add it manually.
        dataset = Dataset(
            data_vars={"variance_density": variance_density, "depth": depth}
        )
        if "time" not in dataset:
            dataset["time"] = to_datetime64(data.time.values)
        return FrequencyDirectionSpectrum(dataset)



from roguewave.wavephysics.balance.dissipation import Dissipation
from typing import Literal
from roguewave.wavephysics.balance.st6_wave_breaking import ST6WaveBreaking
from roguewave.wavephysics.balance.st4_wave_breaking import ST4WaveBreaking
from roguewave.wavephysics.balance.romero_wave_breaking import RomeroWaveBreaking

from roguewave.wavephysics.balance.st4_wind_input import ST4WindInput
from roguewave.wavephysics.balance.st6_wind_input import ST6WindInput

from roguewave.wavephysics.balance.balance import SourceTermBalance

breaking_parametrization = Literal["st6", "st4","romero"]

def create_breaking_dissipation(
    breaking_parametrization: breaking_parametrization = "st6", **kwargs
) -> Dissipation:
    if breaking_parametrization == "st6":
        return ST6WaveBreaking(**kwargs)
    elif breaking_parametrization == "st4":
        return ST4WaveBreaking(**kwargs)
    elif breaking_parametrization == "romero":
        return RomeroWaveBreaking(**kwargs)
    else:
        raise ValueError(f"Unknown parametrization {breaking_parametrization}")


wind_parametrizations = Literal["st6", "st4", "st4_swell"]


def create_wind_source_term(
    wind_parametrization: wind_parametrizations = "st6", **kwargs
):
    if wind_parametrization == "st6":
        return ST6WindInput(**kwargs)
    elif wind_parametrization == "st4":
        return ST4WindInput(**kwargs)
    else:
        raise ValueError(f"Unknown parametrization: {wind_parametrization}")


def create_balance(
    generation_par: wind_parametrizations = "st6",
    dissipation_par: breaking_parametrization = "st6",
    generation_args=None,
    dissipation_args=None,
) -> SourceTermBalance:

    dissipation_args = {} if dissipation_args is None else dissipation_args
    generation_args = {} if generation_args is None else generation_args

    return SourceTermBalance(
        generation=create_wind_source_term(generation_par, **generation_args),
        disspipation=create_breaking_dissipation(dissipation_par, **dissipation_args),
    )



from numpy.fft import rfft, irfft
from numpy import linspace


def upsample(signal, factor: int, t0=0, sampling_frequency=2.5):
    """
    Spectral upsampling. There will be edge effects unless the signal has been windowed.

    :param signal:
    :param factor:
    :return:
    """
    n = len(signal)
    upsampled_time = (
        linspace(0, n * factor, n * factor, endpoint=False)
        / (sampling_frequency * factor)
        + t0
    )

    return upsampled_time, irfft(rfft(signal) * factor, n * factor)



import numpy
from typing import List
from numpy import isnan, ones, abs, sign, median, squeeze, where
from roguewave import FrequencySpectrum, FrequencyDirectionSpectrum, WaveSpectrum
from xarray import DataArray
from roguewave.wavephysics.balance import SourceTermBalance
from roguewave.wavephysics.windestimate import estimate_u10_from_spectrum
from scipy.optimize import minimize


def rmse(target, actual, jacobian_actual=None, weights=None):
    N = len(target)
    if weights is None:
        weights = 1 / N

    mean_error = sum(weights * (actual - target) ** 2) / 2

    if jacobian_actual is not None:
        error_gradient = (weights * (actual - target))[None, :] @ jacobian_actual
        return mean_error, squeeze(error_gradient)
    else:
        return mean_error


def mae(target, actual, jacobian_actual=None, weights=None):
    N = len(target)
    error = sum(abs(actual - target)) / N

    if jacobian_actual is not None:
        error_gradient = (sign(actual - target)[None, :] @ jacobian_actual) / N
        return error, squeeze(error_gradient)
    else:
        return error


def create_weighted_metric(name, binsize, number_of_bins, target):
    bins = numpy.linspace(0, number_of_bins * binsize, number_of_bins, endpoint=False)
    index = numpy.digitize(target, bins)
    weights = numpy.zeros(len(target))
    for ii in range(number_of_bins + 1):
        mask = ii == index
        n = numpy.sum(mask)

        if n == 0:
            continue

        weights[mask] = 1 / (n * number_of_bins)

    return create_metric(name, weights)


def create_metric(name, weights=None):
    if name == "rmse":
        func = rmse
    elif name == "mae":
        func = mae
    elif name == "huber":
        func = huber

    def _closure(target, actual, jacobian_actual=None):
        return func(target, actual, jacobian_actual, weights=weights)

    return _closure


def huber(target, actual, jacobian_actual=None, weights=None):
    diff = actual - target
    delta = abs(diff)
    N = len(target)
    error = sum(where(delta < 1, delta**2 / 2, delta - 1 / 2)) / N

    if jacobian_actual is not None:
        vector = where(delta < 1, diff, sign(diff))
        error_gradient = (vector @ jacobian_actual) / N
        return error, squeeze(error_gradient)
    else:
        return error


def calibrate_wind_estimate_from_spectrum(
    method,
    target_u10: DataArray,
    spectrum: FrequencySpectrum,
    parameter_names: List[str] = None,
    loss_function=None,
    velocity_scale=None,
    bounds=None,
    params=None,
):
    if parameter_names is None:
        parameter_names = [
            "phillips_constant_beta",
            "charnock_constant",
        ]

    estimate_default_parameters = {
        "directional_spreading_constant": 2.5,
        "phillips_constant_beta": 0.012,
        "charnock_constant": 0.015,
    }

    scale = {}

    for parameter_name in parameter_names:
        scale[parameter_name] = estimate_default_parameters[parameter_name]

    if loss_function is None:
        loss_function = rmse

    if velocity_scale is None:
        velocity_scale = median(target_u10)

    x0 = numpy.ones(len(parameter_names))

    def training_function(values):
        estimate_input = {}
        for parameter_name, value in zip(parameter_names, values):
            estimate_input[parameter_name] = value * scale[parameter_name]

        estimate_input = estimate_default_parameters | estimate_input | params

        estimate = estimate_u10_from_spectrum(spectrum, method=method, **estimate_input)

        actual = estimate["u10"].values / velocity_scale
        actual[isnan(actual)] = 0.0
        target = target_u10.values / velocity_scale
        return loss_function(target, actual)

    if bounds is None:
        _bounds = [(0.01, 100) for x in x0]
    else:
        _bounds = []
        for parameter_name in parameter_names:
            x = bounds[parameter_name]
            _scale = scale[parameter_name]
            _bounds.append((x[0] / _scale, x[1] / _scale))

    options = {"maxiter": 100, "disp": True}

    res = minimize(
        training_function, x0, method="L-BFGS-B", bounds=_bounds, options=options
    )

    return {key: x * scale[key] for key, x in zip(parameter_names, res.x)}


def calibrate_wind_estimate_from_balance(
    balance: SourceTermBalance,
    parameter_names: List[str],
    target_u10: DataArray,
    spectrum: FrequencyDirectionSpectrum,
    loss_function=None,
    velocity_scale=None,
    params=None,
    time_derivative_spectrum: FrequencyDirectionSpectrum = None,
):
    dissipation = balance.dissipation
    generation = balance.generation

    if params is not None:
        balance.update_parameters(params)

    scale = {}
    for parameter_name in parameter_names:
        if parameter_name in dissipation._parameters:
            par = dissipation._parameters
        elif parameter_name in generation._parameters:
            par = generation._parameters
        else:
            raise ValueError(f"unknown parameter {parameter_name}")
        scale[parameter_name] = par[parameter_name]

    if loss_function is None:
        loss_function = rmse

    if velocity_scale is None:
        velocity_scale = median(target_u10)

    def training_function(values):
        """
        Training function that returns the current loss value.
        :param values: list of scaled floats, in the order of the names in parameter names
        :return: the loss function and an approximation of the gradient
        """
        params = {}
        for parameter_name, value in zip(parameter_names, values):
            params[parameter_name] = value * scale[parameter_name]

        balance.update_parameters(params)

        # Estimate the wind speed, direction and gradient.
        estimate = balance.windspeed_and_direction_from_spectra(
            target_u10,
            spectrum,
            jacobian=True,
            jacobian_parameters=parameter_names,
            time_derivative_spectrum=time_derivative_spectrum,
        )

        actual = estimate["u10"].values / velocity_scale
        actual[isnan(actual)] = 0.0

        target = target_u10.values / velocity_scale

        jacobian = estimate["jacobian"].values / velocity_scale

        err, grad_err = loss_function(target, actual, jacobian_actual=jacobian)

        # Apply scaling
        for index, parameter_name in enumerate(parameter_names):
            grad_err[index] = grad_err[index] * scale[parameter_name]

        return (err, grad_err)

    x0 = ones(len(parameter_names))
    bounds = [(0.01, 100) for x in x0]
    options = {"maxiter": 100, "disp": True}

    res = minimize(
        training_function,
        x0,
        method="L-BFGS-B",
        bounds=bounds,
        options=options,
        jac=True,
    )

    return {key: x * scale[key] for key, x in zip(parameter_names, res.x)}


def prep_data(
    spectrum: WaveSpectrum, target_u10: DataArray, threshold=(-numpy.inf, numpy.inf)
):
    mask = (
        (spectrum.is_valid())
        & (target_u10.values >= threshold[0])
        & (target_u10.values <= threshold[1])
    )

    return spectrum.where(mask), target_u10[mask.values]



"""
Contents: IO

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
======================

Routines that can be used to save and load data.

Functions:

How To Use This Module
======================
(See the individual functions for details.)

"""
from roguewave.wavespectra import (
    FrequencySpectrum,
    FrequencyDirectionSpectrum,
    create_1d_spectrum,
)
from pandas import DataFrame, read_json
from typing import Union, Dict, List
from datetime import datetime
import gzip
import json
import numpy
import base64
from io import BytesIO
from roguewave import to_datetime_utc, filecache
from xarray import Dataset, open_dataset, DataArray
import boto3
from botocore.errorfactory import ClientError
import tempfile


_UNION = Union[
    FrequencySpectrum,
    FrequencyDirectionSpectrum,
    List[FrequencySpectrum],
    Dict[str, List[FrequencyDirectionSpectrum]],
    Dict[str, List[FrequencySpectrum]],
    List[List[FrequencyDirectionSpectrum]],
    List[List[DataFrame]],
    Dict[int, List[DataFrame]],
    Dict[str, DataFrame],
]


def _b64_encode_numpy(val: numpy.ndarray) -> dict:
    if val.dtype == numpy.float64:
        val = val.astype(dtype=numpy.float32)

    if val.dtype == numpy.int64:
        val = val.astype(dtype=numpy.int32)

    dtype = val.dtype.str
    shape = val.shape
    data = base64.b64encode(val)

    return {
        "__class__": "ndarray",
        "data": {"dtype": dtype, "shape": shape, "data": data.decode("utf8")},
    }


def _b64_decode_numpy(data):
    decoded_bytes = base64.b64decode(bytes(data["data"], encoding="utf-8"))
    dtype = numpy.dtype(data["dtype"])
    array = numpy.frombuffer(decoded_bytes, dtype=dtype)
    return numpy.reshape(array, data["shape"])


def _b64_encode_dataset(val: Dataset, name) -> dict:
    net_cdf = val.to_netcdf(engine="scipy")
    data = base64.b64encode(net_cdf)
    return {"__class__": name, "data": {"data": data.decode("utf8")}}


def _b64_encode_dataarray(val: DataArray, name) -> dict:
    dataset = Dataset()
    dataset = dataset.assign({"dataarray": val})
    net_cdf = dataset.to_netcdf(engine="scipy")
    data = base64.b64encode(net_cdf)
    return {"__class__": name, "data": {"data": data.decode("utf8")}}


def _b64_decode_dataset(data) -> Dataset:
    decoded_bytes = base64.b64decode(bytes(data["data"], encoding="utf-8"))
    with BytesIO(decoded_bytes) as fp:
        dataset = open_dataset(fp, engine="scipy")
    return dataset


def _b64_decode_dataarray(data) -> DataArray:
    decoded_bytes = base64.b64decode(bytes(data["data"], encoding="utf-8"))
    with BytesIO(decoded_bytes) as fp:
        dataset = open_dataset(fp, engine="scipy")
    return dataset["dataarray"]


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, numpy.ndarray):
            return _b64_encode_numpy(obj)
        elif isinstance(obj, numpy.int32):
            return int(obj)
        elif isinstance(obj, numpy.float32):
            return float(obj)
        elif isinstance(obj, numpy.int64):
            return int(obj)
        elif isinstance(obj, numpy.float64):
            return float(obj)
        elif isinstance(obj, Dataset):
            return _b64_encode_dataset(obj, "Dataset")
        elif isinstance(obj, DataArray):
            return _b64_encode_dataarray(obj, "DataArray")
        elif isinstance(obj, FrequencySpectrum):
            return _b64_encode_dataset(obj.dataset, "FrequencySpectrum")
        elif isinstance(obj, FrequencyDirectionSpectrum):
            return _b64_encode_dataset(obj.dataset, "FrequencyDirectionSpectrum")
        elif isinstance(obj, datetime):
            return {"__class__": "datetime", "data": datetime.isoformat(obj)}
        elif isinstance(obj, DataFrame):
            return {"__class__": "DataFrame", "data": obj.to_json(date_unit="s")}
        else:
            return json.JSONEncoder.default(self, obj)


def object_hook(dictionary: dict):
    if "__class__" in dictionary:
        if dictionary["__class__"] == "DataFrame":
            # This is a mess- needs to be refactored.
            df = read_json(dictionary["data"])
            if "timestamp" in df:
                df["timestamp"] = df["timestamp"].apply(lambda x: x.tz_localize("utc"))
            elif "time" in df:
                time = to_datetime_utc(df["time"].values)
                # df["time"] = df["time"].apply(lambda x: x.tz_localize("utc"))
                df["time"] = time
            elif "valid_time" in df:
                df["valid_time"] = df["valid_time"].apply(
                    lambda x: x.tz_localize("utc")
                )
            elif "init_time" in df:
                df["valid_time"] = df["valid_time"].apply(
                    lambda x: x.tz_localize("utc")
                )
            else:
                df.index = [x.tz_localize("utc") for x in df.index]
            return df
        elif dictionary["__class__"] == "WaveSpectrum1D":
            data = dictionary["data"]
            return create_1d_spectrum(
                frequency=data["frequency"],
                variance_density=data["varianceDensity"],
                time=data["timestamp"],
                latitude=data["latitude"],
                longitude=data["longitude"],
                a1=data["a1"],
                b1=data["b1"],
                a2=data["a2"],
                b2=data["b2"],
            )
        elif dictionary["__class__"] == "Dataset":
            return _b64_decode_dataset(dictionary["data"])
        elif dictionary["__class__"] == "DataArray":
            return _b64_decode_dataarray(dictionary["data"])
        elif dictionary["__class__"] == "FrequencySpectrum":
            return FrequencySpectrum(_b64_decode_dataset(dictionary["data"]))
        elif dictionary["__class__"] == "FrequencyDirectionSpectrum":
            return FrequencyDirectionSpectrum(_b64_decode_dataset(dictionary["data"]))
        elif dictionary["__class__"] == "ndarray":
            return _b64_decode_numpy(dictionary["data"])
        elif dictionary["__class__"] == "datetime":
            return datetime.fromisoformat(dictionary["data"])
    else:
        return dictionary


def load(filename: str, force_redownload_if_remote=False) -> _UNION:
    """
    Load spectral data as saved by "save_spectrum" from the given file and
    return a (nested) object. The precise format of the output depends on what
    was saved.

    :param filename: path to file to load. If an s3 uri is given (of the form s3://bucket/key) the remote file is
        downloaded and cached locally. Future requests of the same uri are retrieved from the cache.

    :return:
        - Data in the same form it was saved.
    """

    if filename.startswith("s3://"):
        # Get the file from s3 and cache locally
        if force_redownload_if_remote:
            filecache.delete_files(filename, error_if_not_in_cache=False)
        filename = filecache.filepaths([filename])[0]

    with gzip.open(filename, "rb") as file:
        data = file.read().decode("utf-8")
    return json.loads(data, object_hook=object_hook)


def save(_input: _UNION, filename: str, overwrite=True, s3_overwrite=False):
    """
    Save roguewave data in JSON form compressed with gzip.

    :param _input:
        - Data containing python primitives (dict/list) or any of the RogueWave
          classes.

    :param filename: path to save data. If an s3 uri is given (of the form s3://bucket/key) the  file is
        saved to s3. If a local file already exists with the same name the file is overwritten if overwrite = True
        (default), otherwise an error is raised. If an s3 object with the same uri exists we raise an error (unless
        s3_overwrite = True).

    :param overwrite: Default True. If False an error is raised if a file with the same name already exists. By default
        we simply overwrite under the assumption we are aware of local context.

    :param s3_overwrite: Default False. If False an error is raised if an object with the same uri already exists on s3.
        By default we raise an error as we may clash with keys from others.

    :return: None
    """

    def write(filename, _input):
        with gzip.open(filename, "wt") as file:
            json.dump(_input, file, cls=NumpyEncoder)

    if filename.startswith("s3://"):
        with tempfile.TemporaryFile() as temp_file:
            write(temp_file, _input)
            temp_file.seek(0)

            bucket, key = filename.removeprefix("s3://").split("/", 1)
            s3 = boto3.client("s3")

            try:
                s3.head_object(Bucket=bucket, Key=key)
                if not s3_overwrite:
                    raise FileExistsError(
                        f"Key {key} already exists in bucket {bucket}. To overwrite the file (if"
                        f"desired) set s3_overwrite=True"
                    )
            except ClientError:
                # Not found
                pass

            s3.upload_fileobj(temp_file, bucket, key)

    else:
        write(filename, _input)



from roguewave.wavewatch3 import RestartFileTimeStack, open_restart_file_stack
from .timebase import TimeSlice
from .keygeneration import generate_uris


def open_remote_restart_file(
    variable: str,
    time_slice: TimeSlice,
    model_name: str,
    model_definition_file: str,
    cache=True,
    cache_name: str = None,
) -> RestartFileTimeStack:

    aws_keys = generate_uris(variable, time_slice, model_name)
    return open_restart_file_stack(aws_keys, model_definition_file, cache, cache_name)



"""
Contents: Simple file caching routines that automatically cache remote files
          locally for use.

Copyright (C) 2023
Sofar Ocean Technologies

Authors: Pieter Bart Smit
======================

Functions:
- `colocate_points`, function to colocate two point data sources.

"""
from pandas import DataFrame
from scipy.spatial import KDTree
from datetime import timedelta, datetime, timezone
from numpy import cos, sin, pi, empty, nan, unwrap, mean, dtype, full


def colocate_points(
    data: DataFrame, data_to_colocate: DataFrame, radius: float, time_delta: timedelta
) -> DataFrame:
    """
    Collocate data points in the "data_to_colocate" dataframe to points in the original data frame. Both dataframes are
    expected to contain columns denoted: "longitude","latitude","time" filled with corresponding data. Longitude is
    assumed to be in [-180,180), latitude in [-90,90], time is given as a pandas timestamp with or without timezone
    information.

    The functions returns a dataframe of all data in data_to_colocate that is within the given radius and time_delta of
    datapoints in frame 1. The dataframe index corresponds to the point that was matched in "data" dataframe. If
    multiple points in the "data_to_colocate" fall within the requested bounds, the average of those points is returned
    as the result for datatypes for which average is meaningful. Otherwise data from the first point found is returned.

    :param data: reference data
    :param data_to_colocate: data points we want to match to points in the reference data set
    :param radius: maximum spatial radius (in meters) of points we consider a match.
    :param time_delta: maximum time interval of points we consider a match.
    :return:
    """

    # Create the KDTrees
    tree_data = _construct_tree(data, radius, time_delta)
    tree_to_colocate = _construct_tree(data_to_colocate, radius, time_delta)

    # Setup the output structure
    n = len(data["latitude"])
    output = DataFrame()
    for key in data_to_colocate.columns:
        output[key] = full(n, nan, dtype=data_to_colocate[key].dtype)

    # Query the data tree and return all indices of data close in the other tree.
    nearest_neighbour_indices = tree_data.query_ball_tree(tree_to_colocate, radius)

    for index_a, indices_b in enumerate(nearest_neighbour_indices):
        # any matched?, if not continue (leave data set to nan).

        if len(indices_b) > 0:
            for key in data_to_colocate.columns:
                data_entry = data_to_colocate[key].iloc[indices_b]

                # There may be multiple matched. We only assing a single value as matched.
                # Depending on the data type we will return an average or something else.
                if key == "longitude":
                    data_entry = (
                        mean(unwrap(data_entry.values, discont=180, period=360)) + 180
                    ) % 360 - 180

                elif key == "time":
                    data_entry = data_entry.iloc[0]

                elif data_entry.dtype == dtype("O"):
                    data_entry = data_entry.iloc[0]

                else:
                    data_entry = mean(data_entry.values)

                output.at[index_a, key] = data_entry

    # Drop nan entries (no matches) and return dataframe
    return output.dropna()


def _construct_tree(dataset: DataFrame, radius: float, time_delta: timedelta) -> KDTree:
    """
    Construct a KDTRee based on input data frame. The dataframe must have: latitude, longitude, time as columns. Time
    must be given as a proper datetime object.

    To allow for queries on the tree we convert time to a spatial dimension. To this end we have to couple a spatial
    delta (or distance) to an equivalent time delta, which will be very situational dependent.

    :param dataset: pandas Dataset to load into KDTree
    :param radius: spatial radius for which we want to consider neighbours
    :param time_delta: time delta for which we want to consider neighbours
    :return: scipy KDTree object
    """

    data = empty((4, dataset.shape[0]))

    # Ensure the datetime is timezone aware
    if dataset["time"].dt.tz is not None:
        time = dataset["time"].dt.tz_convert(timezone.utc)

    else:
        time = dataset["time"].dt.tz_localize(timezone.utc)

    # Convert latitude and longitude to radians
    lat = dataset["latitude"] * pi / 180
    lon = dataset["longitude"] * pi / 180

    # Convert lat, lon, time into a 4d spatial tuple.

    # To avoid wrapping complications, we convert the spherical latitude and longitude back to an x,y,z coordinate frame
    # usig WGS84 radius. Note that exact representation is not super critical (spherical vs oblate) - under the
    # assumption that some inaccuracy in nearest neighbour queries based on distance is fine.
    _RADIUS_EARTH = 6371008.8
    data[0, :] = _RADIUS_EARTH * cos(lon) * cos(lat)
    data[1, :] = _RADIUS_EARTH * sin(lon) * cos(lat)
    data[2, :] = _RADIUS_EARTH * sin(lat)

    # Convert time to a spatial dimension using a characteristic velocity. We assume we know a characteristic time
    # and spatial interval within which we consider two points to be matched. To convert time to space we associate with
    # these intervals a characteristic speed.
    time_delta_seconds = (
        time - datetime(1970, 1, 1, tzinfo=timezone.utc)
    ).dt.total_seconds()
    characteristic_speed = radius / time_delta.total_seconds()
    data[3, :] = characteristic_speed * time_delta_seconds

    # Build tree and return
    return KDTree(data.transpose())



import numpy
from datetime import datetime
from multiprocessing.pool import ThreadPool
from typing import Sequence, Union, Tuple, Dict
from tqdm import tqdm
from xarray import Dataset, concat
from roguewave import FrequencyDirectionSpectrum, TrackSet
from roguewave.interpolate.nd_interp import NdInterpolator
from roguewave.tools.time import to_datetime_utc, to_datetime64
from roguewave.wavewatch3.restart_file import RestartFile, MAXIMUM_NUMBER_OF_WORKERS


class RestartFileTimeStack:
    def __init__(self, restart_files: Sequence[RestartFile], parallel=True):
        self._restart_files = restart_files
        self.grid = restart_files[0].grid
        self.depth = restart_files[0].depth
        self._time = to_datetime_utc([x.time for x in restart_files])
        self.parallel = parallel

        # To make the progress bar work we need to store progress across
        # calls somewhere.
        self._progres = {"position": 0, "total": None, "leave": True}

    @property
    def frequency(self) -> numpy.ndarray:
        """
        :return: 1D numpy array of frequencies
        """
        return self.grid.frequencies

    @property
    def direction(self) -> numpy.ndarray:
        """
        :return: 1D numpy array of directions
        """
        return self.grid.directions

    @property
    def latitude(self) -> numpy.ndarray:
        """
        :return: 1D numpy array of latitudes.
        """
        return self.grid.latitude

    @property
    def longitude(self) -> numpy.ndarray:
        """
        :return: 1D numpy array of longitudes.
        """
        return self.grid.longitude

    def coordinates(
        self, index: Union[slice, int, numpy.ndarray]
    ) -> Tuple[Union[float, numpy.ndarray], Union[float, numpy.ndarray]]:
        """
        Return the latitude and longitude as a function of the linear index.
        :param index: linear index
        :return:  ( latitude(s), longitude(s)
        """
        ilon = self.grid.longitude_index(index)
        ilat = self.grid.latitude_index(index)
        return self.latitude[ilat], self.longitude[ilon]

    @property
    def number_of_directions(self) -> int:
        """
        :return: number of directions
        """
        return len(self.direction)

    @property
    def number_of_frequencies(self) -> int:
        """
        :return: number of frequencies
        """
        return len(self.frequency)

    @property
    def number_of_latitudes(self) -> int:
        """
        :return: number of latitudes.
        """
        return len(self.latitude)

    @property
    def number_of_longitudes(self) -> int:
        """
        :return: number of longitudes.
        """
        return len(self.longitude)

    @property
    def number_of_spatial_points(self) -> int:
        """
        :return: Number of spatial points in the restart file. This only counts
            the number of sea points and is *not* equal to
            self.number_of_latitudes * self.number_of_longitudes
            Also referred to as "NSEA" in wavewatch III.
        """
        return self.grid.number_of_spatial_points

    @property
    def number_of_spectral_points(self) -> int:
        """
        :return: number of spectral points.
        """
        return self.number_of_frequencies * self.number_of_directions

    @property
    def time(self) -> Sequence[datetime]:
        return self._time

    def __len__(self):
        return len(self._restart_files)

    def __getitem__(self, nargs) -> FrequencyDirectionSpectrum:
        if len(nargs) == 2:
            time_index, linear_index = nargs
        elif len(nargs) == 3:
            time_index, lat_index, lon_index = nargs
            linear_index = self.grid.index(lat_index, lon_index, valid_only=True)
        else:
            raise ValueError("unexpected number of indices")

        if isinstance(time_index, (int, numpy.integer)):
            time_index = [time_index]

        fancy_index = not isinstance(time_index, slice)
        if fancy_index:
            _input = list(zip(time_index, linear_index))
        else:
            time_index = list(range(*time_index.indices(len(self))))
            _input = [(it, linear_index) for it in time_index]

        def _worker(arg) -> Dataset:
            return self._restart_files[arg[0]][arg[1]].dataset.drop("linear_index")

        self._init_progress_bar(len(_input))
        if self.parallel and len(_input) > 1:
            with ThreadPool(processes=MAXIMUM_NUMBER_OF_WORKERS) as pool:
                data = list(
                    tqdm(
                        pool.imap(_worker, _input),
                        total=self._progres["total"],
                        initial=self._progres["position"],
                        leave=self._progres["leave"],
                    )
                )
        else:
            disable_progress_bar = False
            if len(_input) == 1:
                disable_progress_bar = True

            data = list(
                tqdm(
                    map(_worker, _input),
                    total=self._progres["total"],
                    disable=disable_progress_bar,
                    initial=self._progres["position"],
                    leave=self._progres["leave"],
                )
            )
        self._update_progress_bar(len(_input))

        data = concat(data, dim="time")
        data.coords["time"] = to_datetime64([self.time[it] for it in time_index])
        return FrequencyDirectionSpectrum(data)

    def _init_progress_bar(self, total):
        if self._progres["total"] is None:
            self._progres["total"] = total
            self._progres["position"] = 0
            self._progres["leave"] = True
        else:
            if total + self._progres["position"] == self._progres["total"]:
                self._progres["leave"] = True
            else:
                self._progres["leave"] = False

    def _update_progress_bar(self, number):
        self._progres["position"] += number

        if self._progres["position"] == self._progres["total"]:
            self._progres["total"] = None
            self._progres["position"] = 0

    def interpolate(
        self,
        latitude: Union[numpy.ndarray, float],
        longitude: Union[numpy.ndarray, float],
        time: Union[numpy.ndarray, numpy.datetime64, datetime],
    ) -> FrequencyDirectionSpectrum:
        """
        Extract interpolated spectra at given latitudes and longitudes.
        Input can be either a single latitude and longitude pair, or a
        numpy array of latitudes and longitudes.

        :param latitude: latitudes to get interpolated spectra
        :param longitude: longitudes to get interpolated spectra
        :return: Interpolated spectra. Returned data is of  type float32 and
        has the shape:
                (len(indices),
                number_of_frequencies,
                number_of_directions)
        """
        time = to_datetime64(time)
        points = {
            "time": numpy.atleast_1d(time),
            "latitude": numpy.atleast_1d(latitude),
            "longitude": numpy.atleast_1d(longitude),
        }

        periodic_coordinates = {"longitude": 360}

        def _get_data(indices, idims):
            time_index = indices[0]
            index = self.grid.index(
                latitude_index=indices[1], longitude_index=indices[2]
            )

            output = numpy.zeros(
                (len(index), self.number_of_frequencies, self.number_of_directions)
            )
            mask = index >= 0

            output[mask, :, :] = numpy.squeeze(
                self.__getitem__((time_index[mask], index[mask])).variance_density
            )
            output[~mask, :, :] = numpy.nan
            return output

        data_shape = [
            len(self.time),
            len(self.latitude),
            len(self.longitude),
            self.number_of_frequencies,
            self.number_of_directions,
        ]

        data_coordinates = (
            ("time", to_datetime64(self.time)),
            ("latitude", self.latitude),
            ("longitude", self.longitude),
            ("frequency", self.frequency),
            ("direction", self.direction),
        )

        interpolator = NdInterpolator(
            get_data=_get_data,
            data_coordinates=data_coordinates,
            data_shape=data_shape,
            interp_coord_names=list(points.keys()),
            interp_index_coord_name="time",
            data_periodic_coordinates=periodic_coordinates,
            data_period=None,
            data_discont=None,
        )

        self._init_progress_bar(len(latitude) * 8)
        dataset = interpolator.interpolate(points)

        def _get_depth_data(indices, idims):
            index = self.grid.index(
                latitude_index=indices[0], longitude_index=indices[1]
            )
            output = numpy.zeros(len(index))
            mask = index >= 0
            output[mask] = self.depth[index[mask]]
            output[~mask] = numpy.nan
            return output

        depth_points = {
            "latitude": numpy.atleast_1d(latitude),
            "longitude": numpy.atleast_1d(longitude),
        }
        depth_interpolator = NdInterpolator(
            get_data=_get_depth_data,
            data_coordinates=(
                ("latitude", self.latitude),
                ("longitude", self.longitude),
            ),
            data_shape=[self.number_of_latitudes, self.number_of_longitudes],
            interp_coord_names=list(depth_points.keys()),
            interp_index_coord_name="latitude",
            data_periodic_coordinates=periodic_coordinates,
            data_period=None,
            data_discont=None,
        )

        return FrequencyDirectionSpectrum(
            Dataset(
                data_vars={
                    "variance_density": (("time", "frequency", "direction"), dataset),
                    "depth": (("time",), depth_interpolator.interpolate(depth_points)),
                    "longitude": (("time"), longitude),
                    "latitude": (("time"), latitude),
                },
                coords={
                    "time": time,
                    "frequency": self.frequency,
                    "direction": self.direction,
                },
            )
        )

    def interpolate_tracks(
        self, tracks: TrackSet
    ) -> Dict[str, FrequencyDirectionSpectrum]:
        return {
            _id: self.interpolate(x.latitude, x.longitude, x.time)
            for _id, x in tracks.tracks.items()
        }



from pysofar.sofar import SofarApi
from typing import List, Dict, MutableMapping, Union, Any
from pandas import DataFrame
from numpy import unique, inf, array
from roguewave import (
    FrequencySpectrum,
    create_1d_spectrum,
    to_datetime64,
    to_datetime_utc,
)
from pandas import Timestamp

# SofarAPI instance.
_API = None


# Helper functions
def get_spotter_ids(sofar_api: SofarApi = None) -> List[str]:
    """
    Get a list of Spotter ID's that are available through this account.

    :param sofar_api: valid SofarApi instance.
    :return: List of spotters available through this account.
    """
    if sofar_api is None:
        sofar_api = _get_sofar_api()
    return sofar_api.device_ids


# -----------------------------------------------------------------------------


def _get_sofar_api() -> SofarApi:
    """
    Gets a new sofar API object if requested. Returned object is essentially a
    Singleton class-> next calls will return the stored object instead of
    creating a new class. For module internal use only.

    :return: instantiated SofarApi object
    """
    global _API
    if _API is None:
        _API = SofarApi()
    return _API


# -----------------------------------------------------------------------------


# 6) Helper Functions
# =============================================================================
def _unique_filter(data):
    """
    Filter for dual time entries that occur due to bugs in wavefleet (same
    record returned twice)
    :param data:
    :return:
    """
    if len(data) == 0:
        return data

    # Get time
    if isinstance(data[0], FrequencySpectrum):
        time = array([to_datetime_utc(x.time.values).timestamp() for x in data])
    else:
        time = array([x["time"].timestamp() for x in data])

    # Get indices of unique times
    _, unique_indices = unique(time, return_index=True)

    # Return only unique indices
    return [data[index] for index in unique_indices]


# -----------------------------------------------------------------------------


def _none_filter(data: Dict):
    """
    Filter for the occasional occurance of bad data returned from wavefleet.
    :param data:
    :return:
    """
    return list(
        filter(
            lambda x: (x["latitude"] is not None)
            and (x["longitude"] is not None)
            and (x["timestamp"] is not None),
            data,
        )
    )


# -----------------------------------------------------------------------------


def _get_class(key, data) -> Union[MutableMapping, FrequencySpectrum]:
    MAP_API_TO_MODEL_NAMES = {
        "waves": {"timestamp": "time"},
        "surfaceTemp": {
            "timestamp": "time",
            "degrees": "seaSurfaceTemperature",
        },
        "wind": {
            "timestamp": "time",
            "speed": "windVelocity10Meter",
            "direction": "windDirection10Meter",
        },
        "barometerData": {"timestamp": "time", "value": "seaSurfacePressure"},
        "microphoneData": {"timestamp": "time", "value": "soundPressure"},
        "smartMooringData": {
            "timestamp": "time",
        },
    }

    if key == "frequencyData":
        return create_1d_spectrum(
            frequency=array(data["frequency"]),
            variance_density=array(data["varianceDensity"]),
            time=to_datetime64(data["timestamp"]),
            latitude=array(data["latitude"]),
            longitude=array(data["longitude"]),
            a1=array(data["a1"]),
            b1=array(data["b1"]),
            a2=array(data["a2"]),
            b2=array(data["b2"]),
            dims=("frequency",),
            depth=inf,
        )
    else:
        out = {}
        # Here we postprocess non-spectral data. We do three things:
        # 1) rename variable names to standard names
        # 2) apply any postprocessing if appropriate (e.g. convert datestrings to datetimes)
        # 3) add any desired postprocess variables- this is legacy for peak frequency. We
        #    should not add any variables here anymore.
        if key in MAP_API_TO_MODEL_NAMES:
            # Get the mapping from API variable names to standard variable names.
            mapping = MAP_API_TO_MODEL_NAMES[key]
            for var_key, value in data.items():
                # If the name differs get the correct name, otherwise use the
                # name returned from the API
                target_key = mapping[var_key] if var_key in mapping else var_key

                if target_key == "time":
                    out[target_key] = to_datetime_utc(value)
                else:
                    out[target_key] = value

            if key == "waves":
                out["peakFrequency"] = 1 / out["peakPeriod"]

            return out

        else:
            raise Exception(f"Unknown variable {key}")


def as_dataframe(list_of_dict: List[Dict[str, Any]]) -> DataFrame:
    """
    Convert a list of dictionaries to a dataframe. Each dictionary in the list is assumed
    to have the same set of keys.

    :param list_of_dict:
    :return: dataframe with
    """
    data = {key: [] for key in list_of_dict[0].keys()}
    for dictionary in list_of_dict:
        for key in data:
            data[key].append(dictionary[key])
    data["time"] = [Timestamp(x) for x in data["time"]]
    dataframe = DataFrame.from_dict(data)
    # dataframe.set_index("time", inplace=True)

    return dataframe


# -----------------------------------------------------------------------------



import numpy


def wrapped_difference(
    delta: numpy.ndarray, period=2 * numpy.pi, discont=None
) -> numpy.ndarray:
    """
    Calculate the wrapped difference for a given delta for a periodic variable.
    E.g. if the difference between two angles measured in degrees is 359 we
    map this to -1.

    Per default the output range is set to [-1/2, 1/2] * period so that the
    discontinuous wrapping point is set to 1/2 * period. If desired the
    discontinuity can be mapped anywhere in the 0 to period domain, such that
    the output will be restricted to [discontinuity - period, discontinuity].

    :param delta: periodic variable to map to output domain.
    :param period: period
    :param discont: location of the discontinuity (if None, set to period/2)
    :return: delta in the desired periodic domain.
    """

    if period is None:
        return delta

    if discont is None:
        discont = period / 2

    mask = numpy.isfinite(delta)
    output = numpy.full_like(delta, numpy.nan)
    output[mask] = (delta[mask] + period - discont) % period - period + discont
    return output



from datetime import datetime


def bucket_key(
    variable,
    model_id,
    init: datetime,
    hour: int,
    environment,
    resolution,
    region,
    experiment=None,
):

    if environment == "production":
        bucket = "sofar-wx-data-production"
    elif environment == "staging":
        bucket = "sofar-wx-data-staging"
    elif environment == "dev":
        bucket = "sofar-wx-data-dev"

    init_date = init.strftime("%Y%m%d")
    init_cycle = init.strftime("%H")

    if experiment is not None:
        experiment = f"{experiment}"
    else:
        experiment = ""

    hour = "%03i" % hour
    key = (
        f"{model_id}/"
        f"{init_date}/"
        f"{init_cycle}/"
        f"{resolution}/"
        f"{region}/"
        f"{experiment}/"
        f"{model_id}.{init_date}.{init_cycle}.f{hour}.{variable}.nc"
    )
    return bucket, key



from typing import Dict, Sequence, Tuple, Union
from roguewave.spotterapi import get_spectrum
from roguewave.modeldata.open_remote_restart_files import \
    open_remote_restart_file
from roguewave.modeldata.timebase import TimeSlice
from roguewave.interpolate.geometry import TrackSet
from roguewave.tools.time import to_datetime64
from roguewave import FrequencyDirectionSpectrum
from xarray import Dataset, concat


# =============================================================================
def colocate_model_spotter_spectra(
        variable: str,
        spotter_ids: Sequence[str],
        time_slice: TimeSlice,
        model_name: str,
        model_definition: str,
        cache: bool = True,
        cache_name: str = None,
        spectral_domain: str = 'model',
        timebase: str = 'model',
        return_as_dataset = True
) -> Union[ Tuple[FrequencyDirectionSpectrum,FrequencyDirectionSpectrum]
           ,Tuple[Dict[str,FrequencyDirectionSpectrum],
                  Dict[str,FrequencyDirectionSpectrum]]]:

    if return_as_dataset:
        if (not timebase == 'model') and (len(spotter_ids) > 1):
            raise ValueError('Cannot return as an xarray dataset if the '
                             'time base is native or spotter')

    stack = open_remote_restart_file(variable, time_slice, model_name,
                                     model_definition, cache_name=cache_name,
                                     cache=cache)
    spotters = get_spectrum(spotter_ids, time_slice.start_time,
                            time_slice.end_time, cache=cache)

    if timebase == 'spotter':
        trackset = TrackSet.from_spotters(spotters)
    else:
        trackset = TrackSet.from_spotters(spotters).interpolate(stack.time)

    for spotter_id in spotters:
        spotter_data = spotters[spotter_id].as_frequency_direction_spectrum(
            number_of_directions=stack.number_of_directions
        )

        if timebase == 'model':
            spotter_data = spotter_data.interpolate(
                {'time':to_datetime64(trackset.tracks[spotter_id].time)})

        if spectral_domain == 'model':
            spotter_data =spotter_data.interpolate(
                coordinates={
                    'frequency':stack.frequency,
                     'direction': stack.direction
                }
            )
        spotters[spotter_id] = spotter_data

    print('Get model data along spotter tracks:')
    model = stack.interpolate_tracks(trackset)
    if spectral_domain == 'spotter':
        for spotter_id in model:
            model[spotter_id] = model[spotter_id].interpolate(coordinates={
                    'frequency': spotters['spotter_id']['frequency'],
                     'direction': spotters['spotter_id']['direction']
                })

    if return_as_dataset:
        m = concat( [x.dataset for x in model.values()],'spotter_id')
        model = m.assign_coords(spotter_id=spotter_ids)
        model = FrequencyDirectionSpectrum(model)

        s = concat( [x.dataset for x in spotters.values()],'spotter_id')
        spotters = s.assign_coords(spotter_id=spotter_ids)
        spotters = FrequencyDirectionSpectrum(spotters)

    return model, spotters



from typing import Tuple, Callable, List
import numpy
from roguewave.tools.math import wrapped_difference
from roguewave.tools.grid import enclosing_points_1d
from roguewave.interpolate.general import interpolation_weights_1d


def interpolate_points_nd(
        interpolating_coordinates,
        points,
        periodic_coordinates,
        get_data:Callable[ [List[numpy.ndarray]],numpy.ndarray],
        period_data,
        discont,
        output_shape = None,
        full_coordinates = None
        ):
    """

    :param interpolating_coordinates:
    :param points:
    :param periodic_coordinates:
    :param get_data:
    :param period_data:
    :param discont:
    :param output_shape:
    :param full_coordinates:
    :return:
    """


    number_interp_coor = len(interpolating_coordinates)
    number_points = len(points[interpolating_coordinates[0][0]])

    # Find indices and weights for the succesive 1d interpolation problems
    indices_1d = numpy.empty((number_interp_coor, 2, number_points), dtype='int64')
    weights_1d = numpy.empty((number_interp_coor, 2, number_points), dtype='float64')

    for index, (coordinate_name, coordinate) in enumerate(interpolating_coordinates):
        if coordinate_name in periodic_coordinates:
            period = periodic_coordinates[coordinate_name]
        else:
            period = None

        indices_1d[index, :, :] = enclosing_points_1d(
            coordinate, points[coordinate_name], period=period )
        weights_1d[index, :, :] = interpolation_weights_1d(
            coordinate, points[coordinate_name],
            indices_1d[index, :, :], period=period)

    # We keep a running sum of the weights, if a point is excluded because it
    # contains no data (NaN) the weights will no longer add up to 1 - and we
    # reschale to account for the missing value. This is an easy way to account
    # for interpolation near missing points. Note that if the contribution of
    # missing weights ( 1-weights_sum) exceeds 0.5 - we consider the point
    # invalid.



    if output_shape is None:
        output_shape = (number_points,)
    shape =output_shape
    weights_sum = numpy.zeros(shape)
    if period_data is not None:
        interp_val = numpy.zeros(output_shape, dtype=numpy.complex64)
    else:
        interp_val = numpy.zeros(output_shape, dtype=numpy.float64)

    for intp_indices_nd, intp_weight_nd in _next_point(
            number_interp_coor, indices_1d, weights_1d):
        # Loop over all interpolation points one at a time.
        if period_data is not None:
            to_rad = numpy.pi * 2 / period_data
            val = numpy.exp(1j * get_data(intp_indices_nd) * to_rad)
        else:
            try:
                val = get_data(intp_indices_nd)
            except Exception as e:
                raise e

        mask = numpy.isfinite(val)
        ndim = mask.ndim

        while ndim > 1:
            mask = numpy.all( mask,axis=-1 )
            ndim = mask.ndim



        weights_sum[mask,...] += intp_weight_nd[mask]
        interp_val[mask,...] += intp_weight_nd[mask,...] * val[mask,...]

    interp_val = numpy.where(
        weights_sum > 0.5, interp_val / weights_sum, numpy.nan)

    if period_data is not None:
        to_data_units =  period_data / numpy.pi / 2
        interp_val = wrapped_difference(
            delta=numpy.angle(interp_val) * to_data_units,
            period=period_data,
            discont=discont)

    return interp_val


def _next_point(recursion_depth,
                indices_1d: numpy.ndarray,
                weights_1d: numpy.ndarray,
                *narg) \
        -> Tuple[numpy.ndarray, numpy.ndarray]:
    """
    We are trying to interpolate over N dimensions. In bilinear interpolation
    this means we have to visit 2**N points. If N is known this is most
    clearly expressed as a set of N nested loops:

    J=-1
    for i1 in range(0,2):
        for i2 in range(0,2):
            ...
                for iN in range(0,2):
                    J+=1
                    do stuff for J'th item.

    Here instead, since we do not know N in advance, use a set of recursive
    loops to depth N, where at the final level we yield for each of the 2**N
    points the values of the points and the weights with which they contribute
    to the interpolated value.

    :param recursion_depth:
    :param indices_1d:
    :param weights_1d:
    :param narg: indices from outer recursive loops
    :return: generater function that yields the J"th thing to do stuff with.
    """
    number_of_coordinates = indices_1d.shape[0]
    number_of_points = indices_1d.shape[2]

    if recursion_depth > 0:
        # Loop over the n'th coordinate, with
        #    n = number_of_coordinates - recursion_depth
        for ii in range(0, 2):
            # Yield from next recursive loop, add the loop coordinate to the
            # arg of the next call
            arg = (*narg, ii)
            yield from _next_point(recursion_depth - 1,
                                   indices_1d, weights_1d, *arg)
    else:
        #
        # Here we construct the "fancy" indexes we will use to grab datavalues.
        indices_nd = []
        weights_nd = numpy.ones((number_of_points,), dtype='float64')
        for index in range(0, number_of_coordinates):
            # get the coordinate index for the current point.
            indices_nd.append( indices_1d[index, narg[index], :] )

            # The N-dimensional weight is the multiplication of all weights
            # of the associated 1d problems
            weights_nd *= weights_1d[index, narg[index], :]

        yield indices_nd, weights_nd


from roguewave.wavespectra import (
    FrequencySpectrum,
    create_1d_spectrum,
    concatenate_spectra,
)
from datetime import datetime
from pandas import read_csv, to_numeric
import numpy
import os


def get_spectrum_from_parser_output(path: str) -> FrequencySpectrum:
    """

    :param path: Path that contains the output from the spotter parser.
    :return: A list of WaveSpectrum1D objects.
    """

    def load_spectral_file(file):
        if os.path.isfile(file):
            data = read_csv(file).apply(to_numeric, errors="coerce")
        else:
            raise FileNotFoundError(file)

        columns = list(data.columns)
        frequencies = [float(x) for x in columns[8:]]
        values = data[columns[8:]].values

        time_tuple = data[columns[0:6]].values
        time = []
        for index in range(time_tuple.shape[0]):
            time.append(
                datetime(
                    year=time_tuple[index, 0],
                    month=time_tuple[index, 1],
                    day=time_tuple[index, 2],
                    hour=time_tuple[index, 3],
                    minute=time_tuple[index, 4],
                    second=time_tuple[index, 5],
                )
            )

        return {"time": numpy.array(time), "frequencies": frequencies, "values": values}

    files = ["a1", "b1", "a2", "b2", "Szz"]
    data = {}
    for file_type in files:
        file_location = os.path.join(path, file_type + ".csv")
        data[file_type] = load_spectral_file(file_location)

    number_of_spectra = data["Szz"]["values"].shape[0]
    spectra = []
    for index in range(0, number_of_spectra):
        spectra.append(
            create_1d_spectrum(
                frequency=data["Szz"]["frequencies"],
                variance_density=data["Szz"]["values"][index, :],
                time=data["Szz"]["time"][index],
                latitude=None,
                longitude=None,
                a1=data["a1"]["values"][index, :],
                b1=data["b1"]["values"][index, :],
                a2=data["a2"]["values"][index, :],
                b2=data["b2"]["values"][index, :],
            )
        )
    return concatenate_spectra(spectra, dim="time")



from roguewave import FrequencyDirectionSpectrum
from xarray import DataArray
from numpy import pi, cos, sin, arctan2, empty
from numpy.typing import NDArray
from typing import Literal, Tuple
from roguewave.wavephysics.balance.source_term import SourceTerm
from roguewave.wavespectra.operations import numba_integrate_spectral_data
from roguewave.wavetheory.lineardispersion import inverse_intrinsic_dispersion_relation
from numba import njit, prange


class Dissipation(SourceTerm):
    name = "Dissipation"

    def __init__(self, parameters):
        super(Dissipation, self).__init__(parameters)
        self._dissipation_function = None

    def rate(self, spectrum: FrequencyDirectionSpectrum) -> DataArray:
        data = _dissipation(
            spectrum.variance_density.values,
            spectrum.depth.values,
            self._dissipation_function,
            self.spectral_grid(spectrum),
            self.parameters,
        )
        return DataArray(data=data, coords=spectrum.coords(), dims=spectrum.dims)

    def bulk_rate(self, spectrum: FrequencyDirectionSpectrum) -> DataArray:

        data = _bulk_dissipation(
            spectrum.variance_density.values,
            spectrum.depth.values,
            self._dissipation_function,
            self.spectral_grid(spectrum),
            self.parameters,
        )
        return DataArray(
            data=data, coords=spectrum.coords_space_time, dims=spectrum.dims_space_time
        )

    def mean_direction_degrees(self, spectrum: FrequencyDirectionSpectrum):
        data, _ = _bulk_dissipation_direction(
            spectrum.variance_density.values,
            spectrum.depth.values,
            self._dissipation_function,
            self.spectral_grid(spectrum),
            self.parameters,
        )
        return DataArray(
            data=data, coords=spectrum.coords_space_time, dims=spectrum.dims_space_time
        )


@njit(cache=False, parallel=False)
def _dissipation(
    variance_density, depth, dissipation_source_term_function, spectral_grid, parameters
) -> NDArray:
    (
        number_of_points,
        number_of_frequencies,
        number_of_directions,
    ) = variance_density.shape

    dissipation = empty((number_of_points, number_of_frequencies, number_of_directions))
    for point_index in prange(number_of_points):
        diss = dissipation_source_term_function(
            variance_density[point_index, :, :],
            depth[point_index],
            spectral_grid,
            parameters,
        )
        dissipation[point_index, :, :] = diss
    return dissipation


@njit(cache=False, parallel=False)
def _bulk_dissipation(
    variance_density, depth, dissipation_source_term_function, spectral_grid, parameters
) -> NDArray:
    number_of_points = variance_density.shape[0]
    dissipation = empty((number_of_points))

    for point_index in prange(number_of_points):
        diss = dissipation_source_term_function(
            variance_density=variance_density[point_index, :, :],
            depth=depth[point_index],
            spectral_grid=spectral_grid,
            parameters=parameters,
        )
        dissipation[point_index] = numba_integrate_spectral_data(diss, spectral_grid)
    return dissipation


@njit(cache=False, parallel=False)
def _bulk_dissipation_direction(
    variance_density, depth, dissipation_source_term_function, spectral_grid, parameters
) -> Tuple[NDArray, NDArray]:
    number_of_points = variance_density.shape[0]
    direction = empty((number_of_points))
    bulk = empty((number_of_points))

    for point_index in prange(number_of_points):
        direction[point_index], bulk[point_index] = _bulk_dissipation_direction_point(
            variance_density[point_index, :, :],
            depth[point_index],
            dissipation_source_term_function,
            spectral_grid,
            parameters,
        )

    return direction, bulk


@njit(cache=True)
def _bulk_dissipation_direction_point(
    variance_density, depth, dissipation_source_term_function, spectral_grid, parameters
) -> Tuple[float, float]:
    number_of_frequency = variance_density.shape[0]
    number_of_direction = variance_density.shape[1]

    dissipation = dissipation_source_term_function(
        variance_density=variance_density,
        depth=depth,
        spectral_grid=spectral_grid,
        parameters=parameters,
    )

    bulk = numba_integrate_spectral_data(dissipation, spectral_grid)
    radian_frequency = spectral_grid["radian_frequency"]
    radian_direction = spectral_grid["radian_direction"]
    frequency_step = spectral_grid["frequency_step"]
    direction_step = spectral_grid["direction_step"]
    wave_number = inverse_intrinsic_dispersion_relation(radian_frequency, depth)
    cosine = cos(radian_direction)
    sine = sin(radian_direction)

    kx = 0.0
    ky = 0.0

    # Disspation weighted average wave number to guestimate the wind direction. Note dissipation is negative- hence the
    # minus signs.
    for frequency_index in range(number_of_frequency):
        for direction_index in range(number_of_direction):
            kx -= (
                wave_number[frequency_index]
                * cosine[direction_index]
                * dissipation[frequency_index, direction_index]
                * frequency_step[frequency_index]
                * direction_step[direction_index]
            )

            ky -= (
                wave_number[frequency_index]
                * sine[direction_index]
                * dissipation[frequency_index, direction_index]
                * frequency_step[frequency_index]
                * direction_step[direction_index]
            )

    return (arctan2(ky, kx) * 180 / pi) % 360.0, bulk



import numpy
from numba import njit


@njit(cache=True)
def _fit(x, y, power):
    X = x**power
    coef = numpy.sum(X * y) / numpy.sum(X * X)
    goodness_of_fit = numpy.sum((coef * X - y) ** 2)
    return coef, goodness_of_fit


@njit(cache=True)
def tail_fit(x, y, power):
    if power is None:
        coef_4, goodness_of_fit_4 = _fit(x, y, -4)
        coef_5, goodness_of_fit_5 = _fit(x, y, -5)
        if goodness_of_fit_5 < goodness_of_fit_4:
            return coef_5, -5
        else:
            return coef_4, -4

    else:
        coef, goodness_of_fit = _fit(x, y, power)
        return coef, power


@njit(cache=True)
def fill_zeros_or_nan_in_tail(
    variance_density: numpy.ndarray,
    frequencies: numpy.ndarray,
    power=None,
    zero=0.0,
    points_in_fit=10,
):

    input_shape = variance_density.shape
    number_of_elements = 1
    for value in input_shape[:-1]:
        number_of_elements *= value

    number_of_frequencies = input_shape[-1]
    variance_density = variance_density.reshape((number_of_elements, input_shape[-1]))

    for ii in range(0, number_of_elements):

        for jj in range(number_of_frequencies - 1, -1, -1):
            if variance_density[ii, jj] > zero:
                # Note, for nan values this is also always false. No need to seperately check for that.
                index = jj
                break
        else:
            # no valid value found, we cannot extrapolate. Technically this is not needed as we catch this below as
            # well (since index=number_of_frequencies by default). But to make more explicit that we catch for this
            # scenario I leave it in.
            continue

        if index == number_of_frequencies - 1:
            continue
        elif index < points_in_fit:
            continue

        coef, fitted_power = tail_fit(
            x=frequencies[index - points_in_fit + 1 : index + 1],
            y=variance_density[ii, index - points_in_fit + 1 : index + 1],
            power=power,
        )

        for jj in range(index + 1, number_of_frequencies):
            variance_density[ii, jj] = coef * (frequencies[jj]) ** fitted_power

    return numpy.reshape(variance_density, input_shape)



# Partitioning
from roguewave.wavespectra.partitioning.wavefields import (
    get_bulk_partitions_from_spectral_partitions,
    get_spectral_partitions_from_2dspectra)
from roguewave.wavespectra.partitioning.observations import \
    get_bulk_partitions_from_observations, \
    get_spectral_partitions_from_observations


from numpy import pi, empty
from typing import TypedDict
from roguewave.wavephysics.balance import Dissipation
from roguewave.wavespectra.operations import numba_directionally_integrate_spectral_data
from roguewave.wavetheory.lineardispersion import (
    inverse_intrinsic_dispersion_relation,
    intrinsic_group_velocity,
)
from numba import njit


class ST6WaveBreakingParameters(TypedDict):
    p1: float
    p2: float
    a1: float
    a2: float
    saturation_threshold: float


class ST6WaveBreaking(Dissipation):
    name = "st6 dissipation"

    def __init__(self, parameters: ST6WaveBreakingParameters = None):
        super(ST6WaveBreaking, self).__init__(parameters)
        self._dissipation_function = st6_dissipation

    @staticmethod
    def default_parameters() -> ST6WaveBreakingParameters:
        return ST6WaveBreakingParameters(
            p1=4, p2=4, a1=4.75 * 10**-6, a2=7e-5, saturation_threshold=0.035**2
        )


@njit(cache=True)
def st6_dissipation(variance_density, depth, spectral_grid, parameters):
    number_of_frequencies = variance_density.shape[0]

    radian_frequency = spectral_grid["radian_frequency"]
    saturation_threshold = parameters["saturation_threshold"]

    frequency_spectrum = numba_directionally_integrate_spectral_data(
        variance_density, spectral_grid
    )
    wavenumber = inverse_intrinsic_dispersion_relation(radian_frequency, depth)
    group_velocity = intrinsic_group_velocity(wavenumber, depth)
    saturation_spectrum = frequency_spectrum * group_velocity * wavenumber**3 / 2 / pi

    relative_saturation_exceedence = empty(number_of_frequencies)
    for frequency_index in range(number_of_frequencies):
        relative_exceedence = (
            saturation_spectrum[frequency_index] - saturation_threshold
        ) / saturation_threshold
        if relative_exceedence > 0.0:
            relative_saturation_exceedence[frequency_index] = relative_exceedence
        else:
            relative_saturation_exceedence[frequency_index] = 0.0

    inherent = st6_inherent(
        variance_density, relative_saturation_exceedence, spectral_grid, parameters
    )
    cumulative = st6_cumulative(
        variance_density, relative_saturation_exceedence, spectral_grid, parameters
    )

    return inherent + cumulative


@njit(cache=True)
def st6_inherent(
    variance_density, relative_saturation_exceedence, spectral_grid, parameters
):
    number_of_frequencies = variance_density.shape[0]
    number_of_directions = variance_density.shape[1]

    frequency = spectral_grid["radian_frequency"] / 2.0 / pi
    a1 = parameters["a1"]
    p1 = parameters["p1"]

    inherent = empty((number_of_frequencies, number_of_directions))
    for frequency_index in range(number_of_frequencies):
        for direction_index in range(number_of_directions):
            inherent[frequency_index, direction_index] = (
                -a1
                * relative_saturation_exceedence[frequency_index] ** p1
                * frequency[frequency_index]
                * variance_density[frequency_index, direction_index]
            )
    return inherent


@njit(cache=True)
def st6_cumulative(
    variance_density, relative_saturation_exceedence, spectral_grid, parameters
):
    number_of_frequencies = variance_density.shape[0]
    number_of_directions = variance_density.shape[1]
    frequency_step = spectral_grid["frequency_step"]
    a2 = parameters["a2"]
    p2 = parameters["p2"]

    cumulative = empty((number_of_frequencies, number_of_directions))
    run_sum = 0.0
    for frequency_index in range(number_of_frequencies):
        run_sum += (
            relative_saturation_exceedence[frequency_index]
            * frequency_step[frequency_index]
        )

        for direction_index in range(number_of_directions):
            cumulative[frequency_index, direction_index] = (
                -a2 * run_sum**p2 * variance_density[frequency_index, direction_index]
            )
    return cumulative



from roguewave.wavephysics.fluidproperties import AIR, WATER, GRAVITATIONAL_ACCELERATION
from roguewave.wavespectra.operations import numba_directionally_integrate_spectral_data
from roguewave.wavetheory import (
    inverse_intrinsic_dispersion_relation,
    intrinsic_group_velocity,
)
from roguewave.wavephysics.balance import WindGeneration
from numpy import tanh, cos, pi, sqrt, log, empty, argmax, inf
from numba import njit
from typing import TypedDict


class ST6WaveGenerationParameters(TypedDict):
    gravitational_acceleration: float
    charnock_maximum_roughness: float
    charnock_constant: float
    air_density: float
    water_density: float
    vonkarman_constant: float
    friction_velocity_scaling: float
    elevation: float


class ST6WindInput(WindGeneration):
    name = "st6 generation"

    def __init__(
        self,
        parameters: ST6WaveGenerationParameters = None,
    ):
        super(ST6WindInput, self).__init__(parameters)
        self._wind_source_term_function = _st6_wind_generation_point

    @staticmethod
    def default_parameters() -> ST6WaveGenerationParameters:
        return ST6WaveGenerationParameters(
            gravitational_acceleration=GRAVITATIONAL_ACCELERATION,
            charnock_maximum_roughness=inf,
            charnock_constant=0.015,
            air_density=AIR.density,
            water_density=WATER.density,
            vonkarman_constant=AIR.vonkarman_constant,
            friction_velocity_scaling=28,
            elevation=10,
        )


# ----------------------------------------------------------------------------------------------------------------------
# st6 Point wise implementation functions (apply to a single spatial point)
# ----------------------------------------------------------------------------------------------------------------------


@njit(cache=True)
def _st6_wind_generation_point(
    variance_density,
    wind,
    depth,
    roughness_length,
    st6_spectral_grid,
    st6_parameters,
    wind_source=None,
):
    """
    Implememtation of the st6 wind growth rate term

    :param variance_density:
    :param wind:
    :param depth:
    :param roughness_length:
    :param st6_spectral_grid:
    :param st6_parameters:
    :return:
    """

    # Get the spectral size
    number_of_frequencies, number_of_directions = variance_density.shape

    # Allocate the output variable
    if wind_source is None:
        wind_source = empty((number_of_frequencies, number_of_directions))

    # Unpack variables passed in dictionaries/tuples for ease of reference and some slight speed benifits (avoid dictionary
    # lookup in loops).
    vonkarman_constant = st6_parameters["vonkarman_constant"]
    friction_velocity_scaling = st6_parameters["friction_velocity_scaling"]
    air_density = st6_parameters["air_density"]
    water_density = st6_parameters["water_density"]
    elevation = st6_parameters["elevation"]
    radian_frequency = st6_spectral_grid["radian_frequency"]
    radian_direction = st6_spectral_grid["radian_direction"]
    wind_forcing, wind_direction_degrees, wind_forcing_type = wind

    # Preprocess winds to the correct conventions (U10->friction velocity if need be, wind direction degrees->radians)
    if wind_forcing_type == "u10":
        u10 = wind_forcing
        friction_velocity = u10 * vonkarman_constant / log(elevation / roughness_length)

    elif wind_forcing_type in ["ustar", "friction_velocity"]:
        u10 = wind_forcing / vonkarman_constant * log(elevation / roughness_length)
        friction_velocity = wind_forcing

    else:
        raise ValueError("unknown wind forcing")
    wind_direction_radian = wind_direction_degrees * pi / 180
    us = friction_velocity * friction_velocity_scaling
    frequency_spectrum = numba_directionally_integrate_spectral_data(
        variance_density, st6_spectral_grid
    )
    wavenumber = inverse_intrinsic_dispersion_relation(radian_frequency, depth)
    wavespeed = radian_frequency / wavenumber
    group_velocity = intrinsic_group_velocity(wavenumber, depth)
    saturation_spectrum = frequency_spectrum * group_velocity * wavenumber**3 / 2 / pi

    peak_index = argmax(frequency_spectrum)
    peak_wave_speed = wavespeed[peak_index]
    peak_radian_frequency = radian_frequency[peak_index]

    # precalculate the constant multiplication.
    constant_factor = air_density / water_density

    # Loop over all frequencies/directions
    for frequency_index in range(number_of_frequencies):

        # Since wavenumber and wavespeed only depend on frequency we calculate those outside of the direction loop.
        constant_frequency_factor = constant_factor * radian_frequency[frequency_index]

        st6_directional_spreading_function = 1.12 * (us / peak_wave_speed) ** (-0.5) * (
            radian_frequency[frequency_index] / peak_radian_frequency
        ) ** -(0.95) + 1 / (2 * pi)

        root_of_saturation = sqrt(
            saturation_spectrum[frequency_index] * st6_directional_spreading_function
        )

        for direction_index in range(number_of_directions):
            # Calculate the directional factor in the wind input formulation
            W = (
                u10
                * cos(radian_direction[direction_index] - wind_direction_radian)
                / wavespeed[frequency_index]
                - 1
            )

            if W > 0:
                # If the wave direction has an along wind component we continue.

                directional_saturation = W**2 * root_of_saturation

                st6_sheltering_coefficient = 2.8 - (
                    1 + tanh(10 * directional_saturation - 11)
                )
                growth_rate = (
                    constant_frequency_factor
                    * st6_sheltering_coefficient
                    * directional_saturation
                )

            else:
                growth_rate = 0.0

            wind_source[frequency_index, direction_index] = (
                growth_rate * variance_density[frequency_index, direction_index]
            )
    return wind_source
